import numpy as np
import pytest
from tifffile import tifffile

from nellie.im_info.verifier import FileInfo


class DummyTag:
    def __init__(self, name, value):
        self.name = name
        self.value = value


class DummyVolume:
    def __init__(self, axes_calibration):
        self.axesCalibration = axes_calibration


class DummyChannel:
    def __init__(self, axes_calibration):
        self.volume = DummyVolume(axes_calibration)


class DummyNd2Metadata:
    def __init__(self, axes_calibration=None, channels=None):
        self.volume = DummyVolume(axes_calibration) if axes_calibration is not None else None
        self.channels = channels or []


@pytest.mark.parametrize(
    ("unit", "expected_scale"),
    [
        (tifffile.RESUNIT.CENTIMETER, 1e4),
        (tifffile.RESUNIT.INCH, 25400),
    ],
)
def test_tif_tag_resolution_unit_scaling(tmp_path, unit, expected_scale):
    file_info = FileInfo(str(tmp_path / "dummy.tif"))
    file_info.axes = "YX"
    file_info.dim_res = {"X": None, "Y": None, "Z": None, "T": None}

    metadata = {
        282: DummyTag("XResolution", (2, 1)),
        283: DummyTag("YResolution", (4, 1)),
        296: DummyTag("ResolutionUnit", unit),
    }

    file_info._get_tif_tags_metadata(metadata)

    assert file_info.dim_res["X"] == pytest.approx((1 / 2) * expected_scale)
    assert file_info.dim_res["Y"] == pytest.approx((1 / 4) * expected_scale)


def test_tif_tag_resolution_unit_missing_no_scaling(tmp_path):
    file_info = FileInfo(str(tmp_path / "dummy.tif"))
    file_info.axes = "YX"
    file_info.dim_res = {"X": None, "Y": None, "Z": None, "T": None}

    metadata = {
        282: DummyTag("XResolution", (2, 1)),
        283: DummyTag("YResolution", (4, 1)),
    }

    file_info._get_tif_tags_metadata(metadata)

    assert file_info.dim_res["X"] == pytest.approx(1 / 2)
    assert file_info.dim_res["Y"] == pytest.approx(1 / 4)


def test_nd2_time_increment_uses_median_diff(tmp_path):
    file_info = FileInfo(str(tmp_path / "dummy.nd2"))
    file_info.axes = "TZYX"
    file_info.dim_res = {"X": None, "Y": None, "Z": None, "T": None}

    timestamps = [0.0, 1.0, 2.2, 3.1]
    metadata = {
        "root": DummyNd2Metadata(axes_calibration=[0.2, 0.2, 0.5]),
        "recorded_data": {"Time [s]": timestamps},
    }

    file_info._get_nd2_metadata(metadata)

    assert file_info.dim_res["T"] == pytest.approx(np.median(np.diff(timestamps)))
    assert file_info.dim_res["X"] == pytest.approx(0.2)
    assert file_info.dim_res["Y"] == pytest.approx(0.2)
    assert file_info.dim_res["Z"] == pytest.approx(0.5)


def test_nd2_time_increment_missing_or_single_timepoint(tmp_path):
    file_info = FileInfo(str(tmp_path / "dummy.nd2"))
    file_info.axes = "TYX"
    file_info.dim_res = {"X": None, "Y": None, "Z": None, "T": None}

    metadata = {
        "root": DummyNd2Metadata(axes_calibration=[0.2, 0.2, 0.5]),
        "recorded_data": {"Time [s]": [0.0]},
    }

    file_info._get_nd2_metadata(metadata)

    assert file_info.dim_res["T"] is None
    assert file_info.dim_res["X"] == pytest.approx(0.2)
    assert file_info.dim_res["Y"] == pytest.approx(0.2)
    assert file_info.dim_res["Z"] == pytest.approx(0.5)


def test_nd2_time_increment_missing_time_key(tmp_path):
    file_info = FileInfo(str(tmp_path / "dummy.nd2"))
    file_info.axes = "TYX"
    file_info.dim_res = {"X": None, "Y": None, "Z": None, "T": None}

    metadata = {
        "root": DummyNd2Metadata(axes_calibration=[0.2, 0.2, 0.5]),
        "recorded_data": {},
    }

    file_info._get_nd2_metadata(metadata)

    assert file_info.dim_res["T"] is None
    assert file_info.dim_res["X"] == pytest.approx(0.2)
    assert file_info.dim_res["Y"] == pytest.approx(0.2)
    assert file_info.dim_res["Z"] == pytest.approx(0.5)


def test_nd2_axes_calibration_fallback_to_channel(tmp_path):
    file_info = FileInfo(str(tmp_path / "dummy.nd2"))
    file_info.axes = "ZYX"
    file_info.dim_res = {"X": None, "Y": None, "Z": None, "T": None}

    metadata = {
        "root": DummyNd2Metadata(channels=[DummyChannel([0.1, 0.2, 0.3])]),
        "recorded_data": {},
    }

    file_info._get_nd2_metadata(metadata)

    assert file_info.dim_res["X"] == pytest.approx(0.1)
    assert file_info.dim_res["Y"] == pytest.approx(0.2)
    assert file_info.dim_res["Z"] == pytest.approx(0.3)


def test_change_dim_res_invalid_dimension_raises(tmp_path):
    file_info = FileInfo(str(tmp_path / "dummy.tif"))
    file_info.dim_res = {"X": None, "Y": None, "Z": None, "T": None}

    with pytest.raises(ValueError, match="Invalid dimension"):
        file_info.change_dim_res("Q", 1.0)


def test_select_temporal_range_requires_t_axis(tmp_path):
    file_info = FileInfo(str(tmp_path / "dummy.tif"))
    file_info.axes = "YX"
    file_info.shape = (2, 2)

    with pytest.raises(KeyError, match="No time dimension"):
        file_info.select_temporal_range(0, 1)


def test_validate_preserves_time_range(tmp_path):
    file_info = FileInfo(str(tmp_path / "dummy.tif"))
    file_info.axes = "TYX"
    file_info.shape = (5, 2, 2)
    file_info.dim_res = {"X": 1.0, "Y": 1.0, "Z": None, "T": 1.0}
    file_info.t_start = 1
    file_info.t_end = 3

    file_info._validate()

    assert file_info.t_start == 1
    assert file_info.t_end == 3



import numpy as np

from nellie.feature_extraction.hierarchical import (
    aggregate_stats_for_class,
    append_to_array,
)


class _DummyChild:
    stats_to_aggregate = ["stat"]

    def __init__(self):
        self.stat = [np.array([1.0, 2.0, np.nan, 4.0], dtype=float)]


def test_aggregate_stats_low_memory_parity():
    dummy = _DummyChild()
    list_of_idxs = [np.array([0, 1]), np.array([2, 3])]

    fast = aggregate_stats_for_class(dummy, 0, list_of_idxs, low_memory=False)
    slow = aggregate_stats_for_class(dummy, 0, list_of_idxs, low_memory=True)

    assert fast["stat"]["mean"].shape == slow["stat"]["mean"].shape

    fast_arrs, fast_headers = append_to_array(fast)
    slow_arrs, slow_headers = append_to_array(slow)

    assert fast_headers == slow_headers
    for fast_vals, slow_vals in zip(fast_arrs, slow_arrs):
        assert np.allclose(fast_vals, slow_vals, equal_nan=True)



"""
Network skeletonization and analysis for microscopy images.

This module provides the Network class for skeletonizing network-like structures
and analyzing their topology with optimized CPU/GPU processing.
"""
import itertools
import numpy as np
import skimage.measure
import skimage.morphology as morph
from scipy.spatial import cKDTree
from scipy import ndimage as ndi_cpu

from nellie.utils import adaptive_run
from nellie.utils.base_logger import logger
from nellie.im_info.verifier import ImInfo


class Network:
    """
    Optimized class for analyzing and skeletonizing network-like structures in 3D or 4D microscopy images.

    This version focuses on:
      - Reduced CPU/GPU thrashing.
      - Vectorized neighborhood operations (no Python per-voxel loops on large arrays).
      - More memory-friendly local-max detection.
      - More efficient branch relabeling using distance transforms on per-object crops.
      - Graceful degradation when GPU memory is insufficient (CPU/chunked fallback).

    Parameters
    ----------
    im_info : ImInfo
        Image metadata and paths.
    num_t : int, optional
        Number of timepoints to process. Defaults to all timepoints.
    min_radius_um : float, optional
        Minimum radius of detected objects in micrometers.
    max_radius_um : float, optional
        Maximum radius of detected objects in micrometers.
    viewer : object or None, optional
        Viewer object for status reporting.
    device : {"auto", "cpu", "gpu"}, optional
        Backend selection for connectivity computations.
    low_memory : bool, optional
        If True, use chunked CPU fallbacks for local neighborhood operations.
    max_chunk_voxels : int, optional
        Maximum voxels per chunk for low-memory paths.
    """

    def __init__(
        self,
        im_info: ImInfo,
        num_t=None,
        min_radius_um=0.20,
        max_radius_um=1,
        viewer=None,
        device="auto",
        low_memory: bool = False,
        max_chunk_voxels: int = int(1e6),
    ):

        self.im_info = im_info
        self.device = device
        self.xp, self.ndi, self.device_type = self._resolve_backend(device)
        self.force_device = device is not None and device.lower() in ("cpu", "gpu", "cuda")
        self.low_memory = low_memory
        self.max_chunk_voxels = int(max_chunk_voxels)
        self.num_t = num_t
        if num_t is None and not self.im_info.no_t:
            self.num_t = im_info.shape[im_info.axes.index('T')]

        if not self.im_info.no_z:
            self.z_ratio = self.im_info.dim_res['Z'] / self.im_info.dim_res['X']

        # either (roughly) diffraction limit, or pixel size, whichever is larger
        self.min_radius_um = max(min_radius_um, self.im_info.dim_res['X'])
        self.max_radius_um = max_radius_um

        self.min_radius_px = self.min_radius_um / self.im_info.dim_res['X']
        self.max_radius_px = self.max_radius_um / self.im_info.dim_res['X']

        if self.im_info.no_z:
            self.scaling = (im_info.dim_res['Y'], im_info.dim_res['X'])
        else:
            self.scaling = (im_info.dim_res['Z'], im_info.dim_res['Y'], im_info.dim_res['X'])

        self.shape = ()

        self.im_memmap = None
        self.im_frangi_memmap = None
        self.label_memmap = None
        self.network_memmap = None  # kept for compatibility, not used
        self.pixel_class_memmap = None
        self.skel_memmap = None
        self.skel_relabelled_memmap = None

        self.sigmas = None

        self.debug = None

        self.viewer = viewer

    # -------------------------------------------------------------------------
    # Helper methods for device handling
    # -------------------------------------------------------------------------
    def _resolve_backend(self, device):
        device = (device or "auto").lower()
        if device not in ("auto", "cpu", "gpu", "cuda"):
            raise ValueError(f"Unsupported device '{device}'. Use 'auto', 'cpu', or 'gpu'.")

        if device in ("gpu", "cuda"):
            xp, ndi = self._try_import_cupy(require=True)
            return xp, ndi, "cuda"
        if device == "cpu":
            import numpy as np
            import scipy.ndimage as ndi
            return np, ndi, "cpu"

        xp, ndi = self._try_import_cupy(require=False)
        if xp is not None:
            return xp, ndi, "cuda"
        import numpy as np
        import scipy.ndimage as ndi
        return np, ndi, "cpu"

    def _try_import_cupy(self, require):
        try:
            import cupy
            import cupyx.scipy.ndimage as ndi
        except ModuleNotFoundError as exc:
            if require:
                raise RuntimeError("GPU backend requested but CuPy is not installed.") from exc
            return None, None

        try:
            device_count = cupy.cuda.runtime.getDeviceCount()
        except Exception as exc:
            if require:
                raise RuntimeError("GPU backend requested but CUDA is not available.") from exc
            return None, None

        if device_count <= 0:
            if require:
                raise RuntimeError("GPU backend requested but no CUDA devices were found.")
            return None, None

        return cupy, ndi

    def _is_oom_error(self, exc):
        if isinstance(exc, MemoryError):
            return True
        if self.device_type != "cuda":
            return False
        try:
            import cupy

            return isinstance(exc, cupy.cuda.memory.OutOfMemoryError)
        except Exception:
            return "OutOfMemory" in repr(exc)

    def _free_gpu_memory(self):
        if self.device_type != "cuda":
            return
        try:
            self.xp.get_default_memory_pool().free_all_blocks()
        except Exception:
            return

    def _switch_to_cpu(self):
        import numpy as np
        import scipy.ndimage as ndi

        self.xp = np
        self.ndi = ndi
        self.device_type = "cpu"

    def _set_backend(self, device):
        device = adaptive_run.normalize_device(device)
        self.device = device
        self.xp, self.ndi, self.device_type = self._resolve_backend(device)
        self.force_device = device in ("cpu", "gpu")

    def _set_low_memory(self, low_memory):
        self.low_memory = bool(low_memory)

    def _compute_chunk_shape(self, shape, max_chunk_voxels):
        if max_chunk_voxels is None or max_chunk_voxels <= 0:
            return tuple(shape)
        chunk = list(shape)
        while int(np.prod(chunk)) > max_chunk_voxels:
            idx = int(np.argmax(chunk))
            chunk[idx] = max(1, int(np.ceil(chunk[idx] / 2)))
        return tuple(chunk)

    def _iter_chunks(self, shape, chunk_shape, halo):
        if halo is None or len(halo) != len(shape):
            halo = (0,) * len(shape)
        ranges = [range(0, dim, step) for dim, step in zip(shape, chunk_shape)]
        for starts in itertools.product(*ranges):
            ends = [min(start + step, dim) for start, step, dim in zip(starts, chunk_shape, shape)]
            core = tuple(slice(s, e) for s, e in zip(starts, ends))
            ext_starts = [max(0, s - h) for s, h in zip(starts, halo)]
            ext_ends = [min(dim, e + h) for e, h, dim in zip(ends, halo, shape)]
            ext = tuple(slice(s, e) for s, e in zip(ext_starts, ext_ends))
            core_in_ext = tuple(
                slice(s - es, e - es) for s, e, es in zip(starts, ends, ext_starts)
            )
            yield core, ext, core_in_ext

    def _to_xp(self, arr):
        """
        Convert an array to the backend array type (xp).
        """
        # xp is numpy when running on CPU and cupy when on GPU.
        try:
            return self.xp.asarray(arr)
        except Exception as e:
            if self.device_type == "cuda":
                raise
            logger.warning(f"xp.asarray failed; falling back to numpy. Error: {e}")
            return np.asarray(arr)

    def _to_cpu(self, arr):
        """
        Convert xp array to a numpy array. If already numpy, return as-is.
        """
        if self.device_type == "cuda" and hasattr(arr, "get"):
            return arr.get()
        return np.asarray(arr)

    # -------------------------------------------------------------------------
    # Neighborhood-based skeleton cleanup
    # -------------------------------------------------------------------------
    def _remove_connected_label_pixels(self, skel_labels, force_cpu: bool = False):
        """
        Removes skeleton pixels that are connected to multiple labeled regions.

        This vectorized implementation replaces the original per-pixel Python loop
        with min/max filters over 3x3 (2D) or 3x3x3 (3D) neighborhoods.
        """
        if force_cpu:
            labels_np = np.asarray(skel_labels)
            if self.low_memory:
                return self._remove_connected_label_pixels_chunked(labels_np)
            return self._remove_connected_label_pixels_impl(labels_np, np, ndi_cpu)

        labels_xp = self._to_xp(skel_labels)
        if self.low_memory:
            labels_np = self._to_cpu(labels_xp)
            return self._remove_connected_label_pixels_chunked(labels_np)

        try:
            return self._remove_connected_label_pixels_impl(labels_xp, self.xp, self.ndi)
        except Exception as exc:
            if not self._is_oom_error(exc):
                raise
            self._free_gpu_memory()
            labels_np = self._to_cpu(labels_xp)
            return self._remove_connected_label_pixels_chunked(labels_np)

    def _remove_connected_label_pixels_impl(self, labels, xp, ndi):
        mask = labels > 0

        if self.im_info.no_z:
            size = (3, 3)
        else:
            size = (3, 3, 3)

        max_labels = ndi.maximum_filter(labels, size=size, mode="constant", cval=0)

        bg_val = int(labels.max()) + 1
        labels_no_bg = xp.where(labels == 0, bg_val, labels)
        min_labels = ndi.minimum_filter(labels_no_bg, size=size, mode="constant", cval=bg_val)
        min_labels = xp.where(min_labels == bg_val, 0, min_labels)

        ambiguous = mask & (min_labels > 0) & (max_labels > 0) & (min_labels != max_labels)

        # Preserve original behavior: do not modify boundary voxels.
        boundary = xp.zeros_like(mask, dtype=bool)
        if self.im_info.no_z:
            boundary[0, :] = True
            boundary[-1, :] = True
            boundary[:, 0] = True
            boundary[:, -1] = True
        else:
            boundary[0, :, :] = True
            boundary[-1, :, :] = True
            boundary[:, 0, :] = True
            boundary[:, -1, :] = True
            boundary[:, :, 0] = True
            boundary[:, :, -1] = True

        ambiguous = ambiguous & ~boundary

        cleaned = xp.where(ambiguous, 0, labels)
        return cleaned

    def _remove_connected_label_pixels_chunked(self, labels):
        labels_np = np.asarray(labels)
        shape = labels_np.shape
        halo = (1,) * labels_np.ndim
        chunk_shape = self._compute_chunk_shape(shape, self.max_chunk_voxels)
        cleaned = np.zeros_like(labels_np)

        for core, ext, core_in_ext in self._iter_chunks(shape, chunk_shape, halo):
            chunk = labels_np[ext]
            cleaned_chunk = self._remove_connected_label_pixels_impl(chunk, np, ndi_cpu)
            cleaned[core] = cleaned_chunk[core_in_ext]

        return cleaned

    # -------------------------------------------------------------------------
    # Ensure every object has at least one skeleton voxel
    # -------------------------------------------------------------------------
    def _add_missing_skeleton_labels(self, skel_frame, label_frame, frangi_frame):
        """
        Adds missing labels to the skeleton where the intensity is highest within a labeled region.
        """
        logger.debug("Adding missing skeleton labels.")

        labels_np = np.asarray(label_frame)
        skel_np = np.asarray(skel_frame)
        frangi_np = np.asarray(frangi_frame)

        def _normalize_pos(pos, ndim):
            if pos is None:
                return None
            pos_arr = np.asarray(pos)
            if pos_arr.ndim == 1 and pos_arr.size == ndim:
                return tuple(int(p) for p in pos_arr.tolist())
            if pos_arr.ndim == 2:
                if pos_arr.shape == (1, ndim):
                    return tuple(int(p) for p in pos_arr[0].tolist())
                if pos_arr.shape == (ndim, 1):
                    return tuple(int(p) for p in pos_arr[:, 0].tolist())
            if isinstance(pos, (list, tuple)) and len(pos) == 1:
                inner_arr = np.asarray(pos[0])
                if inner_arr.ndim == 1 and inner_arr.size == ndim:
                    return tuple(int(p) for p in inner_arr.tolist())
            return None

        unique_labels = np.unique(labels_np)
        if unique_labels.size == 0:
            return skel_np
        unique_skel_labels = np.unique(skel_np)

        missing_labels = np.setdiff1d(unique_labels, unique_skel_labels)
        missing_labels = missing_labels[missing_labels != 0]
        if missing_labels.size == 0:
            return skel_np

        try:
            positions = ndi_cpu.maximum_position(
                frangi_np, labels=labels_np, index=missing_labels
            )
        except Exception as exc:
            logger.warning(
                f"Maximum-position lookup failed; leaving {len(missing_labels)} labels without skeletons. "
                f"Error: {exc}"
            )
            return skel_np

        if len(missing_labels) == 1:
            positions = [positions]

        for lab, pos in zip(missing_labels, positions):
            pos = _normalize_pos(pos, skel_np.ndim)
            if pos is None:
                logger.warning(
                    "Skipping missing skeleton label %s due to unrecognized position: "
                    "pos=%s ndim=%s shape=%s",
                    lab,
                    pos,
                    skel_np.ndim,
                    skel_np.shape,
                )
                continue
            if any(p < 0 or p >= dim for p, dim in zip(pos, skel_np.shape)):
                logger.warning(
                    "Skipping missing skeleton label %s due to out-of-bounds position: "
                    "pos=%s shape=%s",
                    lab,
                    pos,
                    skel_np.shape,
                )
                continue
            skel_np[pos] = lab

        return skel_np

    # -------------------------------------------------------------------------
    # Skeletonization
    # -------------------------------------------------------------------------
    def _skeletonize(self, label_frame):
        """
        Skeletonizes the labeled regions on CPU.
        """
        cpu_labels = np.asarray(label_frame)
        if self.low_memory:
            return self._skeletonize_per_object(cpu_labels)

        try:
            skel_mask_cpu = morph.skeletonize(cpu_labels > 0)
        except MemoryError:
            logger.warning("Skeletonization OOM; falling back to per-object skeletonization.")
            return self._skeletonize_per_object(cpu_labels)

        skel_labels_cpu = cpu_labels * skel_mask_cpu
        return skel_labels_cpu

    def _skeletonize_per_object(self, label_frame):
        labels_np = np.asarray(label_frame)
        skel_out = np.zeros_like(labels_np)

        max_label = int(labels_np.max())
        if max_label == 0:
            return skel_out

        slices = ndi_cpu.find_objects(labels_np)
        if slices is None:
            return skel_out

        for lab in range(1, max_label + 1):
            idx = lab - 1
            if idx >= len(slices):
                break
            sl = slices[idx]
            if sl is None:
                continue

            sub_labels = labels_np[sl]
            obj_mask = sub_labels == lab
            if not obj_mask.any():
                continue

            try:
                skel_sub = morph.skeletonize(obj_mask)
            except Exception as exc:
                logger.warning(
                    f"Skeletonization failed for label {lab}; leaving object without skeleton. Error: {exc}"
                )
                continue

            skel_out_sub = skel_out[sl]
            skel_out_sub[skel_sub] = lab
            skel_out[sl] = skel_out_sub

        return skel_out

    # -------------------------------------------------------------------------
    # Sigma management for multi-scale filters
    # -------------------------------------------------------------------------
    def _get_sigma_vec(self, sigma):
        """
        Computes the sigma vector for multi-scale filtering based on image dimensions.
        """
        if self.im_info.no_z:
            sigma_vec = (sigma, sigma)
        else:
            sigma_vec = (sigma / self.z_ratio, sigma, sigma)
        return sigma_vec

    def _set_default_sigmas(self):
        """
        Sets the default sigma values for multi-scale filtering based on the
        minimum and maximum radius in pixels.
        """
        logger.debug('Setting sigma values for multi-scale filters.')
        min_sigma_step_size = 0.2
        num_sigma = 5

        self.sigma_min = self.min_radius_px / 2
        self.sigma_max = self.max_radius_px / 3

        sigma_step_size_calculated = (self.sigma_max - self.sigma_min) / num_sigma
        sigma_step_size = max(min_sigma_step_size, sigma_step_size_calculated)

        # Use numpy here; sigma values are small and do not need GPU.
        self.sigmas = np.arange(self.sigma_min, self.sigma_max, sigma_step_size).tolist()
        logger.debug(f'Calculated sigma step size = {sigma_step_size_calculated}. Sigmas = {self.sigmas}')

    # -------------------------------------------------------------------------
    # Branch relabeling using per-object distance transforms
    # -------------------------------------------------------------------------
    def _relabel_objects(self, branch_skel_labels, label_frame):
        """
        Relabels skeleton pixels by propagating labels to nearby unlabeled pixels.

        This implementation operates per object instance to reduce memory usage.
        For each object label in `label_frame`, it:

          1. Extracts a bounding-box crop containing that object.
          2. Uses the skeleton branch labels inside the crop as seeds.
          3. Runs a distance transform in the crop to find the nearest seed for
             each voxel of the object.
          4. Assigns branch labels to all voxels of the object accordingly.

        Parameters
        ----------
        branch_skel_labels : xp.ndarray or numpy.ndarray
            Branch skeleton labels (non-zero at skeleton voxels).
        label_frame : numpy.ndarray or xp.ndarray
            Instance labels in the image.

        Returns
        -------
        numpy.ndarray
            Relabeled skeleton for the entire frame.
        """
        # Work on CPU for distance transforms; SciPy's EDT is very efficient.
        labels_np = self._to_cpu(label_frame).astype(np.int32, copy=False)
        branch_np = self._to_cpu(branch_skel_labels).astype(np.int32, copy=False)

        relabelled_np = np.zeros_like(labels_np, dtype=np.uint32)

        max_label = int(labels_np.max())
        if max_label == 0:
            return relabelled_np

        # Find object bounding boxes once
        slices = ndi_cpu.find_objects(labels_np)
        if slices is None:
            return relabelled_np

        for lab in range(1, max_label + 1):
            idx = lab - 1
            if idx >= len(slices):
                break
            sl = slices[idx]
            if sl is None:
                continue

            sub_labels = labels_np[sl]
            sub_branch = branch_np[sl]

            obj_mask = (sub_labels == lab)
            if not obj_mask.any():
                continue

            # Seeds are branch labels (>0) inside this object crop
            seed_mask = (sub_branch > 0) & obj_mask
            if not (seed_mask & obj_mask).any():
                # No skeleton seeds for this object; leave unlabeled
                continue

            # For EDT, zeros are considered seeds. We invert the seed mask:
            # seed_mask True -> 0, False -> 1
            edt_input = np.logical_not(seed_mask)

            # Distance transform with indices: returns coordinates of nearest seed voxel.
            # We do not need distances, only indices.
            try:
                indices = ndi_cpu.distance_transform_edt(
                    edt_input,
                    sampling=self.scaling,
                    return_distances=False,
                    return_indices=True,
                )
            except Exception as e:
                logger.warning(
                    f"Distance transform failed for label {lab}. "
                    f"Leaving object unlabeled. Error: {e}"
                )
                continue

            # indices has shape (ndim, ...) and points into sub_branch
            nearest_labels = sub_branch[tuple(indices)]

            # Restrict to the object mask: outside the object remains zero
            nearest_labels[~obj_mask] = 0

            # Merge into global relabelled array.
            relabelled_sub = relabelled_np[sl]
            relabelled_sub[obj_mask] = nearest_labels[obj_mask].astype(np.uint32, copy=False)
            relabelled_np[sl] = relabelled_sub

        return relabelled_np

    # -------------------------------------------------------------------------
    # Multi-scale peak detection with reduced memory footprint
    # -------------------------------------------------------------------------
    def _local_max_peak(self, frame, mask):
        """
        Detects local maxima using multi-scale Laplacian of Gaussian filtering.

        This implementation is memory-friendly: instead of allocating a full
        (num_sigma, *frame.shape) array, it keeps track of the best response
        per voxel across scales.

        Parameters
        ----------
        frame : xp.ndarray or numpy.ndarray
            Input image.
        mask : xp.ndarray or numpy.ndarray
            Binary mask for regions of interest.

        Returns
        -------
        xp.ndarray
            Coordinates of detected local maxima.
        """
        if self.sigmas is None:
            self._set_default_sigmas()

        frame_xp = self._to_xp(frame)
        mask_xp = self._to_xp(mask).astype(bool)

        ndim = frame_xp.ndim
        footprint = self.xp.ones((3,) * ndim)

        best_response = self.xp.zeros_like(frame_xp, dtype=float)
        peak_mask = self.xp.zeros_like(frame_xp, dtype=bool)

        for s in self.sigmas:
            sigma_vec = self._get_sigma_vec(float(s))

            current = -self.ndi.gaussian_laplace(frame_xp, sigma_vec)
            current *= (float(s) ** 2)
            current *= mask_xp
            current = self.xp.where(current < 0, 0, current)

            max_local = self.ndi.maximum_filter(current, footprint=footprint, mode="nearest")
            is_peak = (current == max_local) & (current > best_response) & (current > 0)

            best_response = self.xp.where(is_peak, current, best_response)
            peak_mask = peak_mask | is_peak

        coords_3d = self.xp.argwhere(peak_mask)
        return coords_3d

    # -------------------------------------------------------------------------
    # Skeleton pixel classification
    # -------------------------------------------------------------------------
    def _get_pixel_class(self, skel, force_cpu: bool = False):
        """
        Classifies skeleton pixels into junctions, branches, and endpoints
        based on connectivity.

        Returns
        -------
        xp.ndarray or numpy.ndarray
            Pixel classification:
            0 = background
            1 = isolated pixels
            2 = tips
            3 = edges
            4 = junctions (clipped)
        """
        if force_cpu:
            skel_np = np.asarray(skel)
            if self.low_memory:
                return self._get_pixel_class_chunked(skel_np)
            return self._get_pixel_class_impl(skel_np, np, ndi_cpu)

        skel_xp = self._to_xp(skel)
        if self.low_memory:
            skel_np = self._to_cpu(skel_xp)
            return self._get_pixel_class_chunked(skel_np)

        try:
            return self._get_pixel_class_impl(skel_xp, self.xp, self.ndi)
        except Exception as exc:
            if not self._is_oom_error(exc):
                raise
            self._free_gpu_memory()
            skel_np = self._to_cpu(skel_xp)
            return self._get_pixel_class_chunked(skel_np)

    def _get_pixel_class_impl(self, skel, xp, ndi):
        skel_mask = (skel > 0).astype("uint8")

        if self.im_info.no_z:
            weights = xp.ones((3, 3))
        else:
            weights = xp.ones((3, 3, 3))

        skel_mask_sum = ndi.convolve(skel_mask, weights=weights, mode="constant", cval=0) * skel_mask
        skel_mask_sum[skel_mask_sum > 4] = 4

        return skel_mask_sum

    def _get_pixel_class_chunked(self, skel):
        skel_np = np.asarray(skel)
        shape = skel_np.shape
        halo = (1,) * skel_np.ndim
        chunk_shape = self._compute_chunk_shape(shape, self.max_chunk_voxels)

        if self.im_info.no_z:
            weights = np.ones((3, 3))
        else:
            weights = np.ones((3, 3, 3))

        out = np.zeros_like(skel_np, dtype=np.uint8)
        skel_mask = (skel_np > 0).astype("uint8")

        for core, ext, core_in_ext in self._iter_chunks(shape, chunk_shape, halo):
            chunk = skel_mask[ext]
            chunk_sum = ndi_cpu.convolve(chunk, weights=weights, mode="constant", cval=0)
            core_sum = chunk_sum[core_in_ext] * chunk[core_in_ext]
            core_sum[core_sum > 4] = 4
            out[core] = core_sum.astype(np.uint8, copy=False)

        return out

    # -------------------------------------------------------------------------
    # Time dimension handling
    # -------------------------------------------------------------------------
    def _get_t(self):
        """
        Determines the number of timepoints to process.
        """
        if self.num_t is None:
            if self.im_info.no_t:
                self.num_t = 1
            else:
                self.num_t = self.im_info.shape[self.im_info.axes.index('T')]

    # -------------------------------------------------------------------------
    # Memory allocation for outputs
    # -------------------------------------------------------------------------
    def _allocate_memory(self):
        """
        Allocates memory for skeleton images, pixel classification, and relabeled skeletons.
        """
        logger.debug('Allocating memory for skeletonization.')
        self.label_memmap = self.im_info.get_memmap(self.im_info.pipeline_paths['im_instance_label'])
        self.im_memmap = self.im_info.get_memmap(self.im_info.im_path)
        self.im_frangi_memmap = self.im_info.get_memmap(self.im_info.pipeline_paths['im_preprocessed'])
        self.shape = self.label_memmap.shape

        im_skel_path = self.im_info.pipeline_paths['im_skel']
        self.skel_memmap = self.im_info.allocate_memory(
            im_skel_path,
            dtype='int32',
            description='skeleton image',
            return_memmap=True
        )

        im_pixel_class = self.im_info.pipeline_paths['im_pixel_class']
        self.pixel_class_memmap = self.im_info.allocate_memory(
            im_pixel_class,
            dtype='uint8',
            description='pixel class image',
            return_memmap=True
        )

        im_skel_relabelled = self.im_info.pipeline_paths['im_skel_relabelled']
        self.skel_relabelled_memmap = self.im_info.allocate_memory(
            im_skel_relabelled,
            dtype='uint32',
            description='skeleton relabelled image',
            return_memmap=True
        )

    # -------------------------------------------------------------------------
    # Branch skeleton labels (excluding junctions)
    # -------------------------------------------------------------------------
    def _get_branch_skel_labels(self, pixel_class, force_cpu: bool = False):
        """
        Gets the branch skeleton labels, excluding junctions and background pixels.

        Parameters
        ----------
        pixel_class : xp.ndarray
            Classified skeleton pixels.

        Returns
        -------
        xp.ndarray
            Branch skeleton labels (connected components of non-junction pixels).
        """
        if force_cpu:
            pc_np = np.asarray(pixel_class)
            non_junctions = (pc_np > 0) & (pc_np != 4)
            if self.im_info.no_z:
                structure = np.ones((3, 3))
            else:
                structure = np.ones((3, 3, 3))
            non_junction_labels, _ = ndi_cpu.label(non_junctions, structure=structure)
            return non_junction_labels

        pc_xp = self._to_xp(pixel_class)
        non_junctions = (pc_xp > 0) & (pc_xp != 4)

        if self.im_info.no_z:
            structure = self.xp.ones((3, 3))
        else:
            structure = self.xp.ones((3, 3, 3))

        try:
            non_junction_labels, _ = self.ndi.label(non_junctions, structure=structure)
        except Exception as exc:
            if not self._is_oom_error(exc):
                raise
            self._free_gpu_memory()
            return self._get_branch_skel_labels(self._to_cpu(pc_xp), force_cpu=True)
        return non_junction_labels

    # -------------------------------------------------------------------------
    # Single timepoint processing
    # -------------------------------------------------------------------------
    def _run_frame(self, t):
        """
        Runs skeletonization and network analysis for a single timepoint.

        Parameters
        ----------
        t : int
            Timepoint index.

        Returns
        -------
        tuple
            (branch_skel_labels, pixel_class, branch_labels)
        """
        logger.info(f"Running network analysis, volume {t}/{self.num_t - 1}")

        try:
            return self._run_frame_backend(t)
        except Exception as exc:
            if self.device_type != "cuda" or not self._is_oom_error(exc):
                raise
            logger.warning("GPU OOM in networking; falling back to CPU for this frame.")
            self._free_gpu_memory()
            self._switch_to_cpu()
            return self._run_frame_backend(t)

    def _run_frame_backend(self, t):
        label_frame = self.label_memmap[t]
        label_frame_cpu = np.asarray(label_frame)
        frangi_frame_cpu = np.asarray(self.im_frangi_memmap[t])

        skel_frame = self._skeletonize(label_frame_cpu)
        skel_clean = self._remove_connected_label_pixels(skel_frame, force_cpu=True)
        skel_clean = self._add_missing_skeleton_labels(
            skel_clean, label_frame_cpu, frangi_frame_cpu
        )

        skel_pre_cpu = (skel_clean > 0) * label_frame_cpu

        if self.device_type == "cuda" and not self.low_memory:
            skel_pre = self._to_xp(skel_pre_cpu)
            pixel_class = self._get_pixel_class(skel_pre)
            branch_skel_labels = self._get_branch_skel_labels(pixel_class)
        else:
            pixel_class = self._get_pixel_class(skel_pre_cpu, force_cpu=True)
            branch_skel_labels = self._get_branch_skel_labels(pixel_class, force_cpu=True)

        branch_labels = self._relabel_objects(branch_skel_labels, label_frame_cpu)

        return branch_skel_labels, pixel_class, branch_labels

    # -------------------------------------------------------------------------
    # Optional junction cleanup
    # -------------------------------------------------------------------------
    def _clean_junctions(self, pixel_class):
        """
        Cleans up junctions by removing closely spaced junction pixels.

        This method uses regionprops to group junction pixels and keeps only the
        pixel closest to the junction centroid as a junction, converting the
        others to branch pixels.

        Parameters
        ----------
        pixel_class : numpy.ndarray or xp.ndarray
            Pixel classification of skeleton points.

        Returns
        -------
        numpy.ndarray
            Cleaned pixel classification with redundant junctions removed.
        """
        pc_np = self._to_cpu(pixel_class).copy()

        junctions = pc_np == 4
        if not junctions.any():
            return pc_np

        junction_labels = skimage.measure.label(junctions)
        junction_objects = skimage.measure.regionprops(junction_labels)
        junction_centroids = [obj.centroid for obj in junction_objects]

        for junction_num, junction in enumerate(junction_objects):
            coords = junction.coords
            if len(coords) < 2:
                continue
            # Use KD-tree to find closest pixel to centroid
            junction_tree = cKDTree(coords)
            _, nearest_idx = junction_tree.query(junction_centroids[junction_num], k=1, workers=-1)
            # Convert all other pixels in this junction component to branch class (3)
            coords_list = coords.tolist()
            coords_list.pop(nearest_idx)
            coords_arr = np.array(coords_list).T
            pc_np[tuple(coords_arr)] = 3

        return pc_np

    # -------------------------------------------------------------------------
    # Full networking pipeline
    # -------------------------------------------------------------------------
    def _run_networking(self):
        """
        Runs the network analysis process for all timepoints in the image.
        """
        for t in range(self.num_t):
            if self.viewer is not None:
                self.viewer.status = f'Extracting branches. Frame: {t + 1} of {self.num_t}.'

            skel, pixel_class, skel_relabelled = self._run_frame(t)

            if self.im_info.no_t or self.num_t == 1:
                # Single frame or static image
                if self.device_type == "cuda":
                    self.skel_memmap[:] = self._to_cpu(skel)
                    self.pixel_class_memmap[:] = self._to_cpu(pixel_class)
                    self.skel_relabelled_memmap[:] = self._to_cpu(skel_relabelled)
                else:
                    self.skel_memmap[:] = skel
                    self.pixel_class_memmap[:] = pixel_class
                    self.skel_relabelled_memmap[:] = skel_relabelled
            else:
                # Time series
                if self.device_type == "cuda":
                    self.skel_memmap[t] = self._to_cpu(skel)
                    self.pixel_class_memmap[t] = self._to_cpu(pixel_class)
                    self.skel_relabelled_memmap[t] = self._to_cpu(skel_relabelled)
                else:
                    self.skel_memmap[t] = skel
                    self.pixel_class_memmap[t] = pixel_class
                    self.skel_relabelled_memmap[t] = skel_relabelled

    # -------------------------------------------------------------------------
    # Public entry point
    # -------------------------------------------------------------------------
    def run(self):
        """
        Execute the full network analysis pipeline.
        """
        device = adaptive_run.normalize_device(self.device)
        gpu_ok = adaptive_run.gpu_available()
        if device == "gpu" and not gpu_ok:
            logger.warning("Network: GPU requested but not available; falling back to CPU.")
        if device == "cpu" or not gpu_ok:
            device_order = ["cpu"]
        else:
            device_order = ["gpu", "cpu"]

        start_low_memory = bool(self.low_memory) or adaptive_run.should_use_low_memory(
            self.im_info, include_gpu="gpu" in device_order
        )
        if start_low_memory and not self.low_memory:
            logger.info("Network: enabling low-memory mode based on estimated usage.")

        last_exc = None
        for dev, low in adaptive_run.mode_candidates(device_order, start_low_memory):
            try:
                self._set_backend(dev)
                self._set_low_memory(low)
                self._get_t()
                self._allocate_memory()
                self._run_networking()
                return
            except Exception as exc:
                last_exc = exc
                if adaptive_run.is_gpu_unavailable_error(exc) and dev == "gpu":
                    logger.warning("Network: GPU backend unavailable; retrying on CPU.")
                    continue
                if adaptive_run.is_oom_error(exc):
                    logger.warning(
                        "Network: OOM on %s/%s; retrying with lower settings.",
                        dev,
                        "low-memory" if low else "high-memory",
                    )
                    continue
                raise
        raise last_exc


if __name__ == "__main__":
    im_path = r"D:\\test_files\\nelly_tests\\deskewed-2023-07-13_14-58-28_000_wt_0_acquire.ome.tif"
    im_info = ImInfo(im_path)
    skel = Network(im_info, num_t=3)
    skel.run()



"""
Configures the base logger for the Nellie package.
"""
import logging


logger = logging
logger.basicConfig(
    level=logging.INFO,
    format="%(asctime)s.%(msecs)03d :: %(levelname)s:%(name)s:[%(filename)s:%(lineno)d] :: %(message)s",
    datefmt="%Y-%m-%d | %H:%M:%S",
)
logger.getLogger('xmlschema').setLevel(logger.WARNING)



import os.path
"""
Hierarchical feature extraction for microscopy images.

This module provides the Hierarchy class for extracting multi-level features (voxels, nodes,
branches, organelles, and images) from segmented and tracked microscopy data.
"""
import pickle
import time

import numpy as np
import pandas as pd
from scipy import spatial
from skimage.measure import regionprops

import warnings

warnings.filterwarnings("ignore", category=RuntimeWarning, module="numpy")
warnings.filterwarnings("ignore", category=UserWarning, module="skimage.measure")
warnings.filterwarnings("ignore", category=RuntimeWarning, module="skimage.measure")
warnings.filterwarnings("ignore", category=RuntimeWarning, module="skimage.morphology")
warnings.filterwarnings(
    "ignore",
    category=RuntimeWarning,
    message="Mean of empty slice"
)
warnings.filterwarnings(
    "ignore",
    category=RuntimeWarning,
    message="All-NaN slice encountered"
)
warnings.filterwarnings(
    "ignore",
    category=RuntimeWarning,
    message="invalid value encountered in divide"
)

from nellie.utils import adaptive_run
from nellie.utils.base_logger import logger
from nellie.im_info.verifier import ImInfo
from nellie.tracking.flow_interpolation import FlowInterpolator

# Optional GPU support via CuPy
try:
    import cupy as cp

    _HAS_CUPY = True
except Exception:  # no CuPy or GPU
    cp = None
    _HAS_CUPY = False


class Hierarchy:
    """
    Main orchestration class for hierarchical feature extraction.
    """

    def __init__(
        self,
        im_info: ImInfo,
        skip_nodes: bool = True,
        viewer=None,
        use_gpu: bool = True,
        low_memory: bool = False,
        enable_motility: bool = True,
        enable_adjacency: bool = True,
        device: str | None = None,
        node_chunk_size: int | None = None,
        max_node_mask_elems: int = int(5e7),
    ):
        """
        Parameters
        ----------
        im_info : ImInfo
            Image metadata object.
        skip_nodes : bool
            If True, node-level features are skipped.
        viewer : optional
            Viewer object with `.status` attribute for status updates.
        use_gpu : bool
            If True and CuPy is available, some computations will attempt to use GPU.
            Ignored when device is set to "cpu" or "gpu".
        low_memory : bool
            If True, use low-memory (slower) aggregation strategies where possible.
        enable_motility : bool
            If False, skip all motion-related features (flow interpolation, velocities, etc.).
        enable_adjacency : bool
            If False, skip adjacency map construction.
        device : {"auto", "cpu", "gpu"}, optional
            Backend selection for GPU-eligible operations. "auto" uses GPU if available.
        node_chunk_size : int, optional
            Target number of voxels per chunk when assigning voxels to nodes.
        max_node_mask_elems : int, optional
            Upper bound on the size of the node/voxel mask (num_nodes * chunk_size).
        """
        self.im_info = im_info
        # This may be overwritten in _get_t, but is a good default
        self.num_t = self.im_info.shape[0]

        if self.im_info.no_z:
            self.spacing = (self.im_info.dim_res["Y"], self.im_info.dim_res["X"])
        else:
            self.spacing = (
                self.im_info.dim_res["Z"],
                self.im_info.dim_res["Y"],
                self.im_info.dim_res["X"],
            )

        self.skip_nodes = skip_nodes
        self.viewer = viewer

        self.low_memory = low_memory
        self.enable_motility = enable_motility
        self.enable_adjacency = enable_adjacency
        self.device = (device or "auto").lower()
        self._prefer_gpu = bool(use_gpu)
        self.use_gpu = self._resolve_device(self.device, self._prefer_gpu)
        self.node_chunk_size = node_chunk_size
        self.max_node_mask_elems = int(max_node_mask_elems)

        # Memory-mapped images
        self.im_raw = None
        self.im_struct = None
        self.im_distance = None
        self.im_skel = None
        self.im_pixel_class = None
        self.label_components = None
        self.label_branches = None
        self.im_border_mask = None
        self.im_obj_reassigned = None
        self.im_branch_reassigned = None

        # Flow interpolators (lazy init in run)
        self.flow_interpolator_fw: FlowInterpolator | None = None
        self.flow_interpolator_bw: FlowInterpolator | None = None

        # Hierarchy levels
        self.voxels = None
        self.nodes = None
        self.branches = None
        self.components = None
        self.image = None

    def _cupy_available(self) -> bool:
        if not _HAS_CUPY or cp is None:
            return False
        try:
            return cp.cuda.runtime.getDeviceCount() > 0
        except Exception:
            return False

    def _resolve_device(self, device: str, use_gpu: bool) -> bool:
        if device not in ("auto", "cpu", "gpu", "cuda"):
            raise ValueError(f"Unsupported device '{device}'. Use 'auto', 'cpu', or 'gpu'.")
        if device in ("gpu", "cuda"):
            if not self._cupy_available():
                raise RuntimeError("GPU backend requested but CuPy/CUDA is not available.")
            return True
        if device == "cpu":
            return False
        return bool(use_gpu and self._cupy_available())

    def _set_backend(self, device: str):
        device = adaptive_run.normalize_device(device)
        self.device = device
        self.use_gpu = self._resolve_device(device, self._prefer_gpu)

    def _set_low_memory(self, low_memory):
        self.low_memory = bool(low_memory)

    def _resolve_node_chunk_size(self, num_nodes: int, num_voxels: int) -> int:
        if num_voxels <= 0:
            return 1
        base_chunk = self.node_chunk_size or 10000
        max_mask_elems = self.max_node_mask_elems
        if self.low_memory:
            max_mask_elems = max(1, max_mask_elems // 4)
        if num_nodes > 0 and num_nodes * base_chunk > max_mask_elems:
            base_chunk = max(1, max_mask_elems // num_nodes)
        return int(max(1, min(base_chunk, num_voxels)))

    def _get_t(self) -> int:
        """
        Determine number of time frames.
        """
        if self.num_t is None and not self.im_info.no_t:
            self.num_t = self.im_info.shape[self.im_info.axes.index("T")]
        return self.num_t

    def _allocate_memory(self):
        """
        Load required image data as memory-mapped arrays.
        """
        self.im_raw = self.im_info.get_memmap(self.im_info.im_path)
        self.im_struct = self.im_info.get_memmap(
            self.im_info.pipeline_paths["im_preprocessed"]
        )
        self.im_distance = self.im_info.get_memmap(
            self.im_info.pipeline_paths["im_distance"]
        )
        self.im_skel = self.im_info.get_memmap(
            self.im_info.pipeline_paths["im_skel"]
        )
        self.label_components = self.im_info.get_memmap(
            self.im_info.pipeline_paths["im_instance_label"]
        )
        self.label_branches = self.im_info.get_memmap(
            self.im_info.pipeline_paths["im_skel_relabelled"]
        )
        self.im_border_mask = self.im_info.get_memmap(
            self.im_info.pipeline_paths["im_border"]
        )
        self.im_pixel_class = self.im_info.get_memmap(
            self.im_info.pipeline_paths["im_pixel_class"]
        )

        if not self.im_info.no_t:
            obj_reassigned_path = self.im_info.pipeline_paths.get(
                "im_obj_label_reassigned"
            )
            branch_reassigned_path = self.im_info.pipeline_paths.get(
                "im_branch_label_reassigned"
            )
            if obj_reassigned_path and branch_reassigned_path:
                if os.path.exists(obj_reassigned_path) and os.path.exists(
                    branch_reassigned_path
                ):
                    self.im_obj_reassigned = self.im_info.get_memmap(
                        obj_reassigned_path
                    )
                    self.im_branch_reassigned = self.im_info.get_memmap(
                        branch_reassigned_path
                    )

    def _get_hierarchies(self):
        """
        Run voxel, node, branch, component and image analyses.
        """
        self.voxels = Voxels(self)
        logger.info("Running voxel analysis")
        start = time.time()
        self.voxels.run()
        v_time = time.time() - start

        self.nodes = Nodes(self)
        logger.info("Running node analysis")
        start = time.time()
        self.nodes.run()
        n_time = time.time() - start

        self.branches = Branches(self)
        logger.info("Running branch analysis")
        start = time.time()
        self.branches.run()
        b_time = time.time() - start

        self.components = Components(self)
        logger.info("Running component analysis")
        start = time.time()
        self.components.run()
        c_time = time.time() - start

        self.image = Image(self)
        logger.info("Running image analysis")
        start = time.time()
        self.image.run()
        i_time = time.time() - start

        logger.debug(f"Voxel analysis took {v_time:.3f} seconds")
        logger.debug(f"Node analysis took {n_time:.3f} seconds")
        logger.debug(f"Branch analysis took {b_time:.3f} seconds")
        logger.debug(f"Component analysis took {c_time:.3f} seconds")
        logger.debug(f"Image analysis took {i_time:.3f} seconds")

    # -------------------------------------------------------------------------
    # Streaming feature saving
    # -------------------------------------------------------------------------

    @staticmethod
    def _iter_feature_arrays(level, labels=None):
        """
        Generator that yields (t, frame_array, headers_without_t_label)
        for a given level, without materializing the full feature table.
        """
        all_attr = []

        node_attr = getattr(level, "aggregate_node_metrics", None)
        if node_attr:
            all_attr.append(node_attr)

        voxel_attr = getattr(level, "aggregate_voxel_metrics", None)
        if voxel_attr:
            all_attr.append(voxel_attr)

        branch_attr = getattr(level, "aggregate_branch_metrics", None)
        if branch_attr:
            all_attr.append(branch_attr)

        component_attr = getattr(level, "aggregate_component_metrics", None)
        if component_attr:
            all_attr.append(component_attr)

        inherent_features = getattr(level, "features_to_save", [])
        for feature in inherent_features:
            feature_vals = getattr(level, feature, None)
            if feature_vals is None:
                continue
            if len(feature_vals) == 0:
                continue
            # match aggregate_* structure: list over t of dicts
            all_attr.append([{feature: feature_vals[t]} for t in range(len(feature_vals))])

        if not all_attr:
            return

        num_frames = len(all_attr[0])

        for t in range(num_frames):
            time_dict = {}
            for attr in all_attr:
                time_dict.update(attr[t])

            time_array, new_headers = append_to_array(time_dict)

            if labels is None:
                labels_t = np.arange(len(time_array[0]), dtype=np.int64)
            else:
                labels_t = np.asarray(labels[t])

            t_col = np.full(labels_t.shape[0], t, dtype=np.int64)

            # prepend t and label columns
            time_array.insert(0, labels_t)
            time_array.insert(0, t_col)

            frame_array = np.array(time_array).T
            yield t, frame_array, new_headers

    def _save_dfs(self):
        """
        Save features to CSVs, streaming per-frame to limit peak memory usage.
        """
        if self.viewer is not None:
            self.viewer.status = "Saving features to csv files."

        # Voxels
        voxel_path = self.im_info.pipeline_paths["features_voxels"]
        first = True
        voxel_headers_full = None
        for _, frame_array, headers in self._iter_feature_arrays(self.voxels, labels=None):
            if first:
                voxel_headers_full = ["t", "label"] + headers
                first = False
                mode = "w"
                header_flag = True
            else:
                mode = "a"
                header_flag = False
            df = pd.DataFrame(frame_array, columns=voxel_headers_full)
            df.to_csv(voxel_path, index=False, mode=mode, header=header_flag)

        # Nodes
        if not self.skip_nodes:
            node_path = self.im_info.pipeline_paths["features_nodes"]
            first = True
            node_headers_full = None
            for _, frame_array, headers in self._iter_feature_arrays(
                self.nodes, labels=None
            ):
                if first:
                    node_headers_full = ["t", "label"] + headers
                    first = False
                    mode = "w"
                    header_flag = True
                else:
                    mode = "a"
                    header_flag = False
                df = pd.DataFrame(frame_array, columns=node_headers_full)
                df.to_csv(node_path, index=False, mode=mode, header=header_flag)

        # Branches
        branch_path = self.im_info.pipeline_paths["features_branches"]
        first = True
        branch_headers_full = None
        for _, frame_array, headers in self._iter_feature_arrays(
            self.branches, labels=self.branches.branch_label
        ):
            if first:
                branch_headers_full = ["t", "label"] + headers
                first = False
                mode = "w"
                header_flag = True
            else:
                mode = "a"
                header_flag = False
            df = pd.DataFrame(frame_array, columns=branch_headers_full)
            df.to_csv(branch_path, index=False, mode=mode, header=header_flag)

        # Components (organelles)
        component_path = self.im_info.pipeline_paths["features_organelles"]
        first = True
        component_headers_full = None
        for _, frame_array, headers in self._iter_feature_arrays(
            self.components, labels=self.components.component_label
        ):
            if first:
                component_headers_full = ["t", "label"] + headers
                first = False
                mode = "w"
                header_flag = True
            else:
                mode = "a"
                header_flag = False
            df = pd.DataFrame(frame_array, columns=component_headers_full)
            df.to_csv(component_path, index=False, mode=mode, header=header_flag)

        # Image-level features
        image_path = self.im_info.pipeline_paths["features_image"]
        first = True
        image_headers_full = None
        for _, frame_array, headers in self._iter_feature_arrays(self.image, labels=None):
            if first:
                image_headers_full = ["t", "label"] + headers
                first = False
                mode = "w"
                header_flag = True
            else:
                mode = "a"
                header_flag = False
            df = pd.DataFrame(frame_array, columns=image_headers_full)
            df.to_csv(image_path, index=False, mode=mode, header=header_flag)

    def _save_adjacency_maps(self):
        """
        Construct adjacency maps (as edge lists) for voxels, nodes, branches, and components.
        This version avoids dense matrices to reduce memory usage.
        """
        v_n = []
        v_b = []
        v_o = []

        for t in range(len(self.voxels.time)):
            num_voxels = len(self.voxels.coords[t])

            # voxel -> node adjacency (voxel index, node index)
            if not self.skip_nodes:
                voxel_node_lists = self.voxels.node_labels[t]  # list of arrays per voxel
                edges_vn = []
                for voxel_idx, nodes in enumerate(voxel_node_lists):
                    if nodes is None or len(nodes) == 0:
                        continue
                    for n in nodes:
                        edges_vn.append((voxel_idx, int(n)))
                v_n.append(np.array(edges_vn, dtype=np.int64) if edges_vn else np.zeros((0, 2), dtype=np.int64))

            # voxel -> branch adjacency (voxel index, branch index 0-based)
            branch_labels = np.asarray(self.voxels.branch_labels[t], dtype=np.int64)
            mask = branch_labels > 0
            if np.any(mask):
                rows = np.nonzero(mask)[0]
                cols = branch_labels[mask] - 1
                v_b.append(np.column_stack((rows, cols)))
            else:
                v_b.append(np.zeros((0, 2), dtype=np.int64))

            # voxel -> component adjacency (voxel index, component label)
            component_labels = np.asarray(self.voxels.component_labels[t], dtype=np.int64)
            mask = component_labels > 0
            if np.any(mask):
                rows = np.nonzero(mask)[0]
                cols = component_labels[mask]
                v_o.append(np.column_stack((rows, cols)))
            else:
                v_o.append(np.zeros((0, 2), dtype=np.int64))

        n_b = []
        n_o = []
        if not self.skip_nodes:
            for t in range(len(self.nodes.time)):
                # node -> branch
                node_labels = np.asarray(self.nodes.branch_label[t], dtype=np.int64)
                branch_labels = np.asarray(self.branches.branch_label[t], dtype=np.int64)
                if len(branch_labels) == 0:
                    n_b.append(np.zeros((0, 2), dtype=np.int64))
                else:
                    max_label = int(branch_labels.max())
                    label_to_idx = np.full(max_label + 1, -1, dtype=np.int64)
                    label_to_idx[branch_labels] = np.arange(len(branch_labels), dtype=np.int64)
                    branch_idx = label_to_idx[node_labels]
                    mask = branch_idx >= 0
                    rows = np.nonzero(mask)[0]
                    cols = branch_idx[mask]
                    n_b.append(np.column_stack((rows, cols)))

                # node -> component
                node_comp_labels = np.asarray(self.nodes.component_label[t], dtype=np.int64)
                comp_labels = np.asarray(self.components.component_label[t], dtype=np.int64)
                if len(comp_labels) == 0:
                    n_o.append(np.zeros((0, 2), dtype=np.int64))
                else:
                    max_label_c = int(comp_labels.max())
                    label_to_idx_c = np.full(max_label_c + 1, -1, dtype=np.int64)
                    label_to_idx_c[comp_labels] = np.arange(len(comp_labels), dtype=np.int64)
                    comp_idx = label_to_idx_c[node_comp_labels]
                    mask_c = comp_idx >= 0
                    rows_c = np.nonzero(mask_c)[0]
                    cols_c = comp_idx[mask_c]
                    n_o.append(np.column_stack((rows_c, cols_c)))

        b_o = []
        for t in range(len(self.branches.time)):
            branch_comp_labels = np.asarray(self.branches.component_label[t], dtype=np.int64)
            comp_labels = np.asarray(self.components.component_label[t], dtype=np.int64)
            if len(comp_labels) == 0:
                b_o.append(np.zeros((0, 2), dtype=np.int64))
            else:
                max_label_c = int(comp_labels.max())
                label_to_idx_c = np.full(max_label_c + 1, -1, dtype=np.int64)
                label_to_idx_c[comp_labels] = np.arange(len(comp_labels), dtype=np.int64)
                comp_idx = label_to_idx_c[branch_comp_labels]
                mask_c = comp_idx >= 0
                rows_c = np.nonzero(mask_c)[0]
                cols_c = comp_idx[mask_c]
                b_o.append(np.column_stack((rows_c, cols_c)))

        edges = {
            "v_b": v_b,
            "v_n": v_n,
            "v_o": v_o,
            "n_b": n_b,
            "n_o": n_o,
            "b_o": b_o,
        }

        with open(self.im_info.pipeline_paths["adjacency_maps"], "wb") as f:
            pickle.dump(edges, f)

    def _run_hierarchy(self):
        self._get_t()

        # Lazily initialize flow interpolators only if needed
        if (
            self.enable_motility
            and not self.im_info.no_t
            and self.num_t is not None
            and self.num_t > 1
        ):
            self.flow_interpolator_fw = FlowInterpolator(self.im_info)
            self.flow_interpolator_bw = FlowInterpolator(self.im_info, forward=False)
        else:
            self.flow_interpolator_fw = None
            self.flow_interpolator_bw = None

        self._allocate_memory()
        self._get_hierarchies()
        self._save_dfs()

        if self.viewer is not None:
            self.viewer.status = "Finalizing run."

        if self.enable_adjacency:
            self._save_adjacency_maps()

        if self.viewer is not None:
            self.viewer.status = "Done!"

    def run(self):
        """
        Main execution method.
        """
        device = adaptive_run.normalize_device(self.device)
        gpu_ok = adaptive_run.gpu_available()
        if device == "gpu" and not gpu_ok:
            logger.warning("Hierarchy: GPU requested but not available; falling back to CPU.")
        if device == "auto" and not self._prefer_gpu:
            device_order = ["cpu"]
        elif device == "cpu" or not gpu_ok:
            device_order = ["cpu"]
        else:
            device_order = ["gpu", "cpu"]

        start_low_memory = bool(self.low_memory) or adaptive_run.should_use_low_memory(
            self.im_info, include_gpu="gpu" in device_order
        )
        if start_low_memory and not self.low_memory:
            logger.info("Hierarchy: enabling low-memory mode based on estimated usage.")

        last_exc = None
        for dev, low in adaptive_run.mode_candidates(device_order, start_low_memory):
            try:
                self._set_backend(dev)
                self._set_low_memory(low)
                self._run_hierarchy()
                return
            except Exception as exc:
                last_exc = exc
                if adaptive_run.is_gpu_unavailable_error(exc) and dev == "gpu":
                    logger.warning("Hierarchy: GPU backend unavailable; retrying on CPU.")
                    continue
                if adaptive_run.is_oom_error(exc):
                    logger.warning(
                        "Hierarchy: OOM on %s/%s; retrying with lower settings.",
                        dev,
                        "low-memory" if low else "high-memory",
                    )
                    continue
                raise
        raise last_exc


def append_to_array(to_append):
    """
    Convert feature dict into list-of-arrays + headers.
    """
    new_array = []
    new_headers = []
    for feature, stats in to_append.items():
        if not isinstance(stats, dict):
            stats = {"raw": stats}
            stats["raw"] = [np.array(stats["raw"])]
        for stat, vals in stats.items():
            vals = np.array(vals)[0]
            new_array.append(vals)
            new_headers.append(f"{feature}_{stat}")
    return new_array, new_headers


def create_feature_array(level, labels=None):
    """
    Original non-streaming implementation kept for backwards compatibility.
    Not used inside Hierarchy anymore (which streams to CSV directly).
    """
    full_array = None
    headers = None
    all_attr = []
    attr_dict = []

    if node_attr := getattr(level, "aggregate_node_metrics", None):
        all_attr.append(node_attr)
    if voxel_attr := getattr(level, "aggregate_voxel_metrics", None):
        all_attr.append(voxel_attr)
    if branch_attr := getattr(level, "aggregate_branch_metrics", None):
        all_attr.append(branch_attr)
    if component_attr := getattr(level, "aggregate_component_metrics", None):
        all_attr.append(component_attr)

    inherent_features = getattr(level, "features_to_save", [])
    for feature in inherent_features:
        if feature_vals := getattr(level, feature, None):
            all_attr.append([{feature: feature_vals[t]} for t in range(len(feature_vals))])

    if not all_attr:
        return np.zeros((0, 0)), []

    for t in range(len(all_attr[0])):
        time_dict = {}
        for attr in all_attr:
            time_dict.update(attr[t])
        attr_dict.append(time_dict)

    for t in range(len(attr_dict)):
        to_append = attr_dict[t]
        time_array, new_headers = append_to_array(to_append)
        if labels is None:
            labels_t = np.array(range(len(time_array[0])), dtype=np.int64)
        else:
            labels_t = labels[t]
        time_array.insert(0, labels_t)
        time_array.insert(0, np.array([t] * len(time_array[0]), dtype=np.int64))
        if headers is None:
            headers = new_headers
        if full_array is None:
            full_array = np.array(time_array).T
        else:
            time_array = np.array(time_array).T
            full_array = np.vstack([full_array, time_array])

    headers.insert(0, "label")
    headers.insert(0, "t")
    return full_array, headers


class Voxels:
    """
    Voxel-level features.
    """

    def __init__(self, hierarchy: Hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.coords = []

        # voxel metrics
        self.x = []
        self.y = []
        self.z = []
        self.intensity = []
        self.structure = []

        self.vec01 = []
        self.vec12 = []

        self.angular_acc = []
        self.angular_vel = []
        self.angular_vel_vector = []
        self.linear_acc = []
        self.linear_vel = []
        self.linear_vel_vector = []

        self.rel_angular_acc = []
        self.rel_angular_vel = []
        self.rel_linear_acc = []
        self.rel_linear_vel = []
        self.rel_directionality = []

        self.node_labels = []
        self.branch_labels = []
        self.component_labels = []
        self.image_name = []

        self.node_dim0_lims = []
        self.node_dim1_lims = []
        self.node_dim2_lims = []
        self.node_voxel_idxs = []

        self.stats_to_aggregate = [
            "linear_vel",
            "angular_vel",
            "linear_acc",
            "angular_acc",
            "rel_linear_vel",
            "rel_angular_vel",
            "rel_linear_acc",
            "rel_angular_acc",
            "rel_directionality",
            "structure",
            "intensity",
        ]

        self.features_to_save = self.stats_to_aggregate + ["x", "y", "z"]

    def _get_node_info(self, t, frame_coords):
        """
        Compute node bounding boxes and voxel->node assignments.
        """
        skeleton_pixels = np.argwhere(self.hierarchy.im_pixel_class[t] > 0)
        skeleton_radius = self.hierarchy.im_distance[t][tuple(skeleton_pixels.T)]

        lims_dim0 = (
            skeleton_radius[:, np.newaxis] * np.array([-1, 1]) + skeleton_pixels[:, 0, np.newaxis]
        ).astype(int)
        lims_dim0[:, 1] += 1
        lims_dim1 = (
            skeleton_radius[:, np.newaxis] * np.array([-1, 1]) + skeleton_pixels[:, 1, np.newaxis]
        ).astype(int)
        lims_dim1[:, 1] += 1

        lims_dim0[lims_dim0 < 0] = 0
        lims_dim1[lims_dim1 < 0] = 0

        if not self.hierarchy.im_info.no_z:
            lims_dim2 = (
                skeleton_radius[:, np.newaxis] * np.array([-1, 1])
                + skeleton_pixels[:, 2, np.newaxis]
            ).astype(int)
            lims_dim2[:, 1] += 1
            lims_dim2[lims_dim2 < 0] = 0
            max_dim0 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index("Z")]
            max_dim1 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index("Y")]
            max_dim2 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index("X")]
            lims_dim2[lims_dim2 > max_dim2] = max_dim2
        else:
            lims_dim2 = None
            max_dim0 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index("Y")]
            max_dim1 = self.hierarchy.im_info.shape[self.hierarchy.im_info.axes.index("X")]

        lims_dim0[lims_dim0 > max_dim0] = max_dim0
        lims_dim1[lims_dim1 > max_dim1] = max_dim1

        self.node_dim0_lims.append(lims_dim0)
        self.node_dim1_lims.append(lims_dim1)
        self.node_dim2_lims.append(lims_dim2)

        frame_coords = np.array(frame_coords)
        num_nodes = len(skeleton_pixels)
        num_voxels = len(frame_coords)
        chunk_size = self.hierarchy._resolve_node_chunk_size(num_nodes, num_voxels)

        def _process_chunks(local_chunk_size):
            chunk_node_voxel_idxs = {idx: [] for idx in range(num_nodes)}
            chunk_nodes_idxs = []
            if num_voxels == 0:
                return chunk_nodes_idxs, chunk_node_voxel_idxs
            num_chunks = int(np.ceil(num_voxels / local_chunk_size))
            for chunk_num in range(num_chunks):
                logger.debug(f"Processing chunk {chunk_num + 1} of {num_chunks}")
                start = chunk_num * local_chunk_size
                end = min((chunk_num + 1) * local_chunk_size, num_voxels)
                chunk_frame_coords = frame_coords[start:end]

                if not self.hierarchy.im_info.no_z:
                    dim0_coords, dim1_coords, dim2_coords = (
                        chunk_frame_coords[:, 0],
                        chunk_frame_coords[:, 1],
                        chunk_frame_coords[:, 2],
                    )
                    dim0_mask = (lims_dim0[:, 0][:, None] <= dim0_coords) & (
                        lims_dim0[:, 1][:, None] >= dim0_coords
                    )
                    dim1_mask = (lims_dim1[:, 0][:, None] <= dim1_coords) & (
                        lims_dim1[:, 1][:, None] >= dim1_coords
                    )
                    dim2_mask = (lims_dim2[:, 0][:, None] <= dim2_coords) & (
                        lims_dim2[:, 1][:, None] >= dim2_coords
                    )
                    mask = dim0_mask & dim1_mask & dim2_mask
                else:
                    dim0_coords, dim1_coords = chunk_frame_coords[:, 0], chunk_frame_coords[:, 1]
                    dim0_mask = (lims_dim0[:, 0][:, None] <= dim0_coords) & (
                        lims_dim0[:, 1][:, None] >= dim0_coords
                    )
                    dim1_mask = (lims_dim1[:, 0][:, None] <= dim1_coords) & (
                        lims_dim1[:, 1][:, None] >= dim1_coords
                    )
                    mask = dim0_mask & dim1_mask

                frame_coord_nodes_idxs = [[] for _ in range(mask.shape[1])]
                rows, cols = np.nonzero(mask)
                for row, col in zip(rows, cols):
                    frame_coord_nodes_idxs[col].append(row)
                frame_coord_nodes_idxs = [
                    np.array(indices) for indices in frame_coord_nodes_idxs
                ]

                chunk_nodes_idxs.extend(frame_coord_nodes_idxs)

                for i in range(num_nodes):
                    chunk_node_voxel_idxs[i].extend(np.nonzero(mask[i])[0] + start)
            return chunk_nodes_idxs, chunk_node_voxel_idxs

        while True:
            try:
                chunk_nodes_idxs, chunk_node_voxel_idxs = _process_chunks(chunk_size)
                break
            except MemoryError:
                if chunk_size <= 1:
                    raise
                logger.warning(
                    "Node assignment OOM with chunk_size=%d; retrying smaller chunks.",
                    chunk_size,
                )
                chunk_size = max(1, chunk_size // 2)

        self.node_labels.append(chunk_nodes_idxs)
        chunk_node_voxel_idxs = [
            np.array(chunk_node_voxel_idxs[i]) for i in range(len(skeleton_pixels))
        ]
        self.node_voxel_idxs.append(chunk_node_voxel_idxs)

    def _get_min_euc_dist(self, t, vec):
        """
        Compute index of voxel with minimal Euclidean distance for each branch label.
        Returns a numpy array of length max_label+1 where element [label] is the
        index into vec/coords for that representative voxel (or NaN if none).
        """
        euc_dist = np.linalg.norm(vec, axis=1)
        branch_labels = np.asarray(self.branch_labels[t], dtype=np.int64)

        if branch_labels.size == 0:
            return np.array([], dtype=float)

        max_label = int(branch_labels.max())
        idxmin = np.full(max_label + 1, np.nan, dtype=float)

        unique_labels = np.unique(branch_labels)
        for lbl in unique_labels:
            mask = branch_labels == lbl
            vals = euc_dist[mask]
            valid = ~np.isnan(vals)
            if not np.any(valid):
                continue
            local_idx = np.nanargmin(vals[valid])
            global_indices = np.nonzero(mask)[0][valid]
            idxmin[lbl] = global_indices[local_idx]

        return idxmin

    def _get_ref_coords(self, coords_a, coords_b, idxmin, t):
        """
        Map every voxel to reference coordinates based on branch label.
        """
        branch_labels = np.asarray(self.branch_labels[t], dtype=np.int64)
        max_label = len(idxmin) - 1
        branch_labels_clipped = np.clip(branch_labels, 0, max_label)
        vals_a = idxmin[branch_labels_clipped]
        vals_b = idxmin[branch_labels_clipped]

        vals_a_no_nan = vals_a.copy()
        vals_a_no_nan[np.isnan(vals_a_no_nan)] = 0
        vals_a_no_nan = vals_a_no_nan.astype(int)

        vals_b_no_nan = vals_b.copy()
        vals_b_no_nan[np.isnan(vals_b_no_nan)] = 0
        vals_b_no_nan = vals_b_no_nan.astype(int)

        ref_a = coords_a[vals_a_no_nan]
        ref_b = coords_b[vals_b_no_nan]

        ref_a[np.isnan(vals_a)] = np.nan
        ref_b[np.isnan(vals_b)] = np.nan

        return ref_a, ref_b

    def _get_linear_velocity(self, ra, rb):
        lin_disp = rb - ra
        dt = self.hierarchy.im_info.dim_res["T"]
        lin_vel = lin_disp / dt
        lin_vel_mag = np.linalg.norm(lin_vel, axis=1)
        # orientation not used downstream; don't return it to save memory
        return lin_vel, lin_vel_mag, None

    def _get_angular_velocity_2d(self, ra, rb):
        theta_a = np.arctan2(ra[:, 1], ra[:, 0])
        theta_b = np.arctan2(rb[:, 1], rb[:, 0])
        delta_theta = theta_b - theta_a
        delta_theta = (delta_theta + np.pi) % (2 * np.pi) - np.pi
        dt = self.hierarchy.im_info.dim_res["T"]
        ang_vel = delta_theta / dt
        ang_vel_mag = np.abs(ang_vel)
        ang_vel_orient = np.sign(ang_vel)
        return ang_vel, ang_vel_mag, ang_vel_orient

    def _get_angular_velocity_3d(self, ra, rb):
        cross_product = np.cross(ra, rb, axis=1)
        norm = np.linalg.norm(ra, axis=1) * np.linalg.norm(rb, axis=1)
        ang_disp = np.divide(cross_product.T, norm.T).T
        ang_disp[norm == 0] = [np.nan, np.nan, np.nan]

        dt = self.hierarchy.im_info.dim_res["T"]
        ang_vel = ang_disp / dt
        ang_vel_mag = np.linalg.norm(ang_vel, axis=1)
        ang_vel_orient = (ang_vel.T / ang_vel_mag).T
        ang_vel_orient = np.where(
            np.isinf(ang_vel_orient),
            [np.nan, np.nan, np.nan],
            ang_vel_orient,
        )
        return ang_vel, ang_vel_mag, ang_vel_orient

    def _get_angular_velocity(self, ra, rb):
        if self.hierarchy.im_info.no_z:
            return self._get_angular_velocity_2d(ra, rb)
        return self._get_angular_velocity_3d(ra, rb)

    def _get_motility_stats(self, t, coords_1_px):
        """
        Compute motility-related features. Can be skipped entirely by setting
        hierarchy.enable_motility = False.
        """
        coords_1_px = coords_1_px.astype("float32")
        dims = 2 if self.hierarchy.im_info.no_z else 3
        n = len(coords_1_px)

        # If motility is disabled or we don't have flows, fill with NaNs and return.
        if (
            not self.hierarchy.enable_motility
            or self.hierarchy.flow_interpolator_fw is None
            or self.hierarchy.flow_interpolator_bw is None
            or self.hierarchy.num_t is None
            or self.hierarchy.num_t < 2
        ):
            self.vec01.append(np.full((n, dims), np.nan, dtype=np.float32))
            self.vec12.append(np.full((n, dims), np.nan, dtype=np.float32))
            nan_vec = np.full((n, dims), np.nan, dtype=np.float32)
            nan_arr = np.full(n, np.nan, dtype=np.float32)

            self.linear_vel_vector.append(nan_vec)
            self.linear_vel.append(nan_arr)
            self.angular_vel_vector.append(nan_arr if dims == 2 else nan_vec)
            self.angular_vel.append(nan_arr)
            self.rel_linear_vel.append(nan_arr)
            self.rel_angular_vel.append(nan_arr)
            self.rel_directionality.append(nan_arr)
            self.linear_acc.append(nan_arr)
            self.angular_acc.append(nan_arr)
            self.rel_linear_acc.append(nan_arr)
            self.rel_angular_acc.append(nan_arr)
            return

        vec01 = []
        vec12 = []

        if t > 0:
            vec01_px = self.hierarchy.flow_interpolator_bw.interpolate_coord(coords_1_px, t)
            vec01 = vec01_px * self.hierarchy.spacing
            self.vec01.append(vec01.astype(np.float32))
        else:
            self.vec01.append(np.full((n, dims), np.nan, dtype=np.float32))

        if t < self.hierarchy.num_t - 1:
            vec12_px = self.hierarchy.flow_interpolator_fw.interpolate_coord(coords_1_px, t)
            vec12 = vec12_px * self.hierarchy.spacing
            self.vec12.append(vec12.astype(np.float32))
        else:
            self.vec12.append(np.full((n, dims), np.nan, dtype=np.float32))

        coords_1 = coords_1_px * self.hierarchy.spacing

        # Forward and backward in time
        if len(vec01) and len(vec12):
            coords_0_px = coords_1_px - vec01_px
            coords_0 = coords_0_px * self.hierarchy.spacing

            lin_vel_01, lin_vel_mag_01, _ = self._get_linear_velocity(coords_0, coords_1)
            ang_vel_01, ang_vel_mag_01, _ = self._get_angular_velocity(coords_0, coords_1)

            idxmin01 = self._get_min_euc_dist(t, vec01)
            ref_coords01 = self._get_ref_coords(coords_0, coords_1, idxmin01, t)
            ref_coords01[0][np.isnan(vec01)] = np.nan
            ref_coords01[1][np.isnan(vec01)] = np.nan
            r0_rel_01 = coords_0 - ref_coords01[0]
            r1_rel_01 = coords_1 - ref_coords01[1]

            lin_vel_rel_01, lin_vel_mag_rel_01, _ = self._get_linear_velocity(
                r0_rel_01, r1_rel_01
            )
            ang_vel_rel_01, ang_vel_mag_rel_01, _ = self._get_angular_velocity(
                r0_rel_01, r1_rel_01
            )

        if len(vec12):
            coords_2_px = coords_1_px + vec12_px
            coords_2 = coords_2_px * self.hierarchy.spacing

            lin_vel, lin_vel_mag, _ = self._get_linear_velocity(coords_1, coords_2)
            ang_vel, ang_vel_mag, _ = self._get_angular_velocity(coords_1, coords_2)

            idxmin12 = self._get_min_euc_dist(t, vec12)
            ref_coords12 = self._get_ref_coords(coords_1, coords_2, idxmin12, t)
            ref_coords12[0][np.isnan(vec12)] = np.nan
            ref_coords12[1][np.isnan(vec12)] = np.nan
            r1_rel_12 = coords_1 - ref_coords12[0]
            r2_rel_12 = coords_2 - ref_coords12[1]

            lin_vel_rel, lin_vel_mag_rel, _ = self._get_linear_velocity(
                r1_rel_12, r2_rel_12
            )
            ang_vel_rel, ang_vel_mag_rel, _ = self._get_angular_velocity(
                r1_rel_12, r2_rel_12
            )

            r2_rel_mag_12 = np.linalg.norm(r2_rel_12, axis=1)
            r1_rel_mag_12 = np.linalg.norm(r1_rel_12, axis=1)
            denom = r2_rel_mag_12 + r1_rel_mag_12
            directionality_rel = np.divide(
                np.abs(r2_rel_mag_12 - r1_rel_mag_12),
                denom,
                out=np.full_like(denom, np.nan, dtype=np.float32),
                where=denom != 0,
            )
        else:
            lin_vel = np.full((n, dims), np.nan, dtype=np.float32)
            lin_vel_mag = np.full(n, np.nan, dtype=np.float32)
            ang_vel_mag = np.full(n, np.nan, dtype=np.float32)
            lin_vel_rel = np.full((n, dims), np.nan, dtype=np.float32)
            lin_vel_mag_rel = np.full(n, np.nan, dtype=np.float32)
            ang_vel_mag_rel = np.full(n, np.nan, dtype=np.float32)
            directionality_rel = np.full(n, np.nan, dtype=np.float32)
            if dims == 3:
                ang_vel = np.full((n, dims), np.nan, dtype=np.float32)
                ang_vel_rel = np.full((n, dims), np.nan, dtype=np.float32)
            else:
                ang_vel = np.full(n, np.nan, dtype=np.float32)
                ang_vel_rel = np.full(n, np.nan, dtype=np.float32)

        self.linear_vel_vector.append(lin_vel.astype(np.float32))
        self.linear_vel.append(lin_vel_mag.astype(np.float32))
        self.angular_vel_vector.append(ang_vel.astype(np.float32))
        self.angular_vel.append(ang_vel_mag.astype(np.float32))
        self.rel_linear_vel.append(lin_vel_mag_rel.astype(np.float32))
        self.rel_angular_vel.append(ang_vel_mag_rel.astype(np.float32))
        self.rel_directionality.append(directionality_rel.astype(np.float32))

        if len(vec01) and len(vec12):
            dt = self.hierarchy.im_info.dim_res["T"]
            lin_acc = (lin_vel - lin_vel_01) / dt
            lin_acc_mag = np.linalg.norm(lin_acc, axis=1)
            ang_acc = (ang_vel - ang_vel_01) / dt

            lin_acc_rel = (lin_vel_rel - lin_vel_rel_01) / dt
            lin_acc_rel_mag = np.linalg.norm(lin_acc_rel, axis=1)

            ang_acc_rel = (ang_vel_rel - ang_vel_rel_01) / dt
            if self.hierarchy.im_info.no_z:
                ang_acc_mag = np.abs(ang_acc)
                ang_acc_rel_mag = np.abs(ang_acc_rel)
            else:
                ang_acc_mag = np.linalg.norm(ang_acc, axis=1)
                ang_acc_rel_mag = np.linalg.norm(ang_acc_rel, axis=1)
        else:
            lin_acc_mag = np.full(n, np.nan, dtype=np.float32)
            ang_acc_mag = np.full(n, np.nan, dtype=np.float32)
            lin_acc_rel_mag = np.full(n, np.nan, dtype=np.float32)
            ang_acc_rel_mag = np.full(n, np.nan, dtype=np.float32)

        self.linear_acc.append(lin_acc_mag.astype(np.float32))
        self.angular_acc.append(ang_acc_mag.astype(np.float32))
        self.rel_linear_acc.append(lin_acc_rel_mag.astype(np.float32))
        self.rel_angular_acc.append(ang_acc_rel_mag.astype(np.float32))

    def _run_frame(self, t=None):
        frame_coords = np.argwhere(self.hierarchy.label_components[t] > 0)
        self.coords.append(frame_coords)

        frame_component_labels = self.hierarchy.label_components[t][tuple(frame_coords.T)]
        self.component_labels.append(frame_component_labels)

        frame_branch_labels = self.hierarchy.label_branches[t][tuple(frame_coords.T)]
        self.branch_labels.append(frame_branch_labels)

        frame_intensity_vals = self.hierarchy.im_raw[t][tuple(frame_coords.T)]
        self.intensity.append(frame_intensity_vals)

        if not self.hierarchy.im_info.no_z:
            frame_z_vals = frame_coords[:, 0]
            frame_y_vals = frame_coords[:, 1]
            frame_x_vals = frame_coords[:, 2]
        else:
            frame_z_vals = np.full(len(frame_coords), np.nan)
            frame_y_vals = frame_coords[:, 0]
            frame_x_vals = frame_coords[:, 1]
        self.z.append(frame_z_vals)
        self.y.append(frame_y_vals)
        self.x.append(frame_x_vals)

        frame_structure_vals = self.hierarchy.im_struct[t][tuple(frame_coords.T)]
        self.structure.append(frame_structure_vals)

        frame_t = np.ones(frame_coords.shape[0], dtype=int) * t
        self.time.append(frame_t)

        im_name = (
            np.ones(frame_coords.shape[0], dtype=object)
            * self.hierarchy.im_info.file_info.filename_no_ext
        )
        self.image_name.append(im_name)

        if not self.hierarchy.skip_nodes:
            self._get_node_info(t, frame_coords)

        self._get_motility_stats(t, frame_coords)

    def run(self):
        if self.hierarchy.num_t is None:
            self.hierarchy.num_t = 1
        for t in range(self.hierarchy.num_t):
            if self.hierarchy.viewer is not None:
                self.hierarchy.viewer.status = (
                    f"Extracting voxel features. Frame: {t + 1} of {self.hierarchy.num_t}."
                )
            self._run_frame(t)


def aggregate_stats_for_class(child_class, t, list_of_idxs, low_memory: bool = False):
    """
    Aggregate mean/std/min/max/sum over groups of indices for a given class at time t.

    Parameters
    ----------
    child_class : object with stats_to_aggregate and per-stat per-frame arrays
    t : int
    list_of_idxs : list of 1D index arrays (groups)
    low_memory : bool
        If True, use a slower but memory-light implementation.
    """
    aggregate_stats = {
        stat_name: {"mean": [], "std_dev": [], "min": [], "max": [], "sum": []}
        for stat_name in child_class.stats_to_aggregate
        if stat_name != "reassigned_label"
    }

    if low_memory:
        # Loop over groups and stats to avoid building large intermediate arrays.
        for stat_name in child_class.stats_to_aggregate:
            if stat_name == "reassigned_label":
                continue

            stat_array = np.array(getattr(child_class, stat_name)[t])

            # Skip multi-dimensional stats as in original implementation
            if stat_array.ndim > 1:
                continue

            for idxs in list_of_idxs:
                if len(idxs) == 0:
                    aggregate_stats[stat_name]["mean"].append(np.nan)
                    aggregate_stats[stat_name]["std_dev"].append(np.nan)
                    aggregate_stats[stat_name]["min"].append(np.nan)
                    aggregate_stats[stat_name]["max"].append(np.nan)
                    aggregate_stats[stat_name]["sum"].append(np.nan)
                else:
                    vals = stat_array[idxs.astype(int)]
                    aggregate_stats[stat_name]["mean"].append(np.nanmean(vals))
                    aggregate_stats[stat_name]["std_dev"].append(np.nanstd(vals))
                    aggregate_stats[stat_name]["min"].append(np.nanmin(vals))
                    aggregate_stats[stat_name]["max"].append(np.nanmax(vals))
                    aggregate_stats[stat_name]["sum"].append(np.nansum(vals))

        # Convert lists to arrays
        for stat_name in aggregate_stats:
            for key in aggregate_stats[stat_name]:
                arr = np.asarray(aggregate_stats[stat_name][key])
                if arr.ndim == 0:
                    arr = arr.reshape(1, 1)
                elif arr.ndim == 1:
                    arr = arr[None, :]
                aggregate_stats[stat_name][key] = arr
        return aggregate_stats

    # Original vectorized implementation (faster but more memory-intensive)
    aggregate_stats = {
        stat_name: {"mean": [], "std_dev": [], "min": [], "max": [], "sum": []}
        for stat_name in child_class.stats_to_aggregate
        if stat_name != "reassigned_label"
    }

    largest_idx = max((len(idxs) for idxs in list_of_idxs), default=0)

    for stat_name in child_class.stats_to_aggregate:
        if stat_name == "reassigned_label":
            continue

        stat_array = np.array(getattr(child_class, stat_name)[t])

        if stat_array.ndim > 1:
            # skip multi-dimensional stats, as in original code
            continue

        # append a NaN at the end
        stat_array = np.append(stat_array, np.nan)

        # Build big index array
        idxs_array = np.full(
            (len(list_of_idxs), largest_idx), len(stat_array) - 1, dtype=int
        )
        for i, idxs in enumerate(list_of_idxs):
            if len(idxs) > 0:
                idxs_array[i, : len(idxs)] = idxs

        stat_values = stat_array[idxs_array.astype(int)]
        if stat_values.shape[1] == 0:
            stat_values = np.full((stat_values.shape[0], 1), np.nan)

        mean = np.nanmean(stat_values, axis=1)
        std_dev = np.nanstd(stat_values, axis=1)
        min_val = np.nanmin(stat_values, axis=1)
        max_val = np.nanmax(stat_values, axis=1)
        sum_val = np.nansum(stat_values, axis=1)

        aggregate_stats[stat_name]["mean"].append(mean)
        aggregate_stats[stat_name]["std_dev"].append(std_dev)
        aggregate_stats[stat_name]["min"].append(min_val)
        aggregate_stats[stat_name]["max"].append(max_val)
        aggregate_stats[stat_name]["sum"].append(sum_val)

    # Convert lists-of-arrays to arrays-of-arrays
    for stat_name in aggregate_stats:
        for key in aggregate_stats[stat_name]:
            aggregate_stats[stat_name][key] = np.array(aggregate_stats[stat_name][key])

    return aggregate_stats


class Nodes:
    """
    Node-level features.
    """

    def __init__(self, hierarchy: Hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.nodes = []

        self.aggregate_voxel_metrics = []

        self.z = []
        self.x = []
        self.y = []
        self.node_thickness = []
        self.divergence = []
        self.convergence = []
        self.vergere = []

        self.stats_to_aggregate = [
            "divergence",
            "convergence",
            "vergere",
            "node_thickness",
        ]

        self.features_to_save = self.stats_to_aggregate + ["x", "y", "z"]

        self.voxel_idxs = self.hierarchy.voxels.node_voxel_idxs
        self.branch_label = []
        self.component_label = []
        self.image_name = []

        self.node_z_lims = self.hierarchy.voxels.node_dim0_lims
        self.node_y_lims = self.hierarchy.voxels.node_dim1_lims
        self.node_x_lims = self.hierarchy.voxels.node_dim2_lims

    def _get_aggregate_voxel_stats(self, t):
        frame_agg = aggregate_stats_for_class(
            self.hierarchy.voxels,
            t,
            self.hierarchy.voxels.node_voxel_idxs[t],
            low_memory=self.hierarchy.low_memory,
        )
        self.aggregate_voxel_metrics.append(frame_agg)

    def _get_node_stats(self, t):
        radius = distance_check(
            self.hierarchy.im_border_mask[t],
            self.nodes[t],
            self.hierarchy.spacing,
        )
        self.node_thickness.append(radius * 2)

        divergence = []
        convergence = []
        vergere = []
        z = []
        y = []
        x = []
        for i, node in enumerate(self.nodes[t]):
            vox_idxs = self.voxel_idxs[t][i]
            if len(vox_idxs) == 0:
                divergence.append(np.nan)
                convergence.append(np.nan)
                vergere.append(np.nan)
                z.append(np.nan)
                y.append(np.nan)
                x.append(np.nan)
                continue

            coords_vox = self.hierarchy.voxels.coords[t][vox_idxs]
            if not self.hierarchy.im_info.no_z:
                z.append(
                    np.nanmean(coords_vox[:, 0]) * self.hierarchy.spacing[0]
                )
                y.append(
                    np.nanmean(coords_vox[:, 1]) * self.hierarchy.spacing[1]
                )
                x.append(
                    np.nanmean(coords_vox[:, 2]) * self.hierarchy.spacing[2]
                )
            else:
                z.append(np.nan)
                y.append(
                    np.nanmean(coords_vox[:, 0]) * self.hierarchy.spacing[0]
                )
                x.append(
                    np.nanmean(coords_vox[:, 1]) * self.hierarchy.spacing[1]
                )

            dist_vox_node = (coords_vox - self.nodes[t][i]).astype(float, copy=False)
            dist_vox_node_mag = np.linalg.norm(dist_vox_node, axis=1, keepdims=True)
            dir_vox_node = np.divide(
                dist_vox_node,
                dist_vox_node_mag,
                out=np.full_like(dist_vox_node, np.nan),
                where=dist_vox_node_mag != 0,
            )

            vec01 = self.hierarchy.voxels.vec01[t][vox_idxs]
            vec12 = self.hierarchy.voxels.vec12[t][vox_idxs]

            dot_prod_01 = -np.nanmean(np.sum(-vec01 * dir_vox_node, axis=1))
            convergence.append(dot_prod_01)

            dot_prod_12 = np.nanmean(np.sum(vec12 * dir_vox_node, axis=1))
            divergence.append(dot_prod_12)

            vergere.append(dot_prod_01 + dot_prod_12)

        self.divergence.append(divergence)
        self.convergence.append(convergence)
        self.vergere.append(vergere)
        self.z.append(z)
        self.y.append(y)
        self.x.append(x)

    def _run_frame(self, t):
        frame_skel_coords = np.argwhere(self.hierarchy.im_pixel_class[t] > 0)
        self.nodes.append(frame_skel_coords)

        frame_t = np.ones(frame_skel_coords.shape[0], dtype=int) * t
        self.time.append(frame_t)

        frame_component_label = self.hierarchy.label_components[t][
            tuple(frame_skel_coords.T)
        ]
        self.component_label.append(frame_component_label)

        frame_branch_label = self.hierarchy.label_branches[t][
            tuple(frame_skel_coords.T)
        ]
        self.branch_label.append(frame_branch_label)

        im_name = (
            np.ones(frame_skel_coords.shape[0], dtype=object)
            * self.hierarchy.im_info.file_info.filename_no_ext
        )
        self.image_name.append(im_name)

        self._get_aggregate_voxel_stats(t)
        self._get_node_stats(t)

    def run(self):
        if self.hierarchy.skip_nodes:
            return
        for t in range(self.hierarchy.num_t):
            if self.hierarchy.viewer is not None:
                self.hierarchy.viewer.status = (
                    f"Extracting node features. Frame: {t + 1} of {self.hierarchy.num_t}."
                )
            self._run_frame(t)


def distance_check(border_mask, check_coords, spacing):
    """
    Compute distance from points in check_coords to nearest border point.
    """
    border_coords = np.argwhere(border_mask) * spacing
    if border_coords.size == 0:
        return np.full(len(check_coords), np.nan, dtype=float)
    border_tree = spatial.cKDTree(border_coords)
    dist, _ = border_tree.query(check_coords * spacing, k=1)
    return dist


class Branches:
    """
    Branch-level features.
    """

    def __init__(self, hierarchy: Hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.branch_label = []

        self.aggregate_voxel_metrics = []
        self.aggregate_node_metrics = []

        self.z = []
        self.y = []
        self.x = []
        self.branch_length = []
        self.branch_thickness = []
        self.branch_aspect_ratio = []
        self.branch_tortuosity = []
        self.branch_area = []
        self.branch_axis_length_maj = []
        self.branch_axis_length_min = []
        self.branch_extent = []
        self.branch_solidity = []
        self.reassigned_label = []

        self.branch_idxs = []
        self.component_label = []
        self.image_name = []

        self.stats_to_aggregate = [
            "branch_length",
            "branch_thickness",
            "branch_aspect_ratio",
            "branch_tortuosity",
            "branch_area",
            "branch_axis_length_maj",
            "branch_axis_length_min",
            "branch_extent",
            "branch_solidity",
            "reassigned_label",
        ]

        self.features_to_save = self.stats_to_aggregate + ["x", "y", "z"]

    def _get_aggregate_stats(self, t):
        voxel_labels = self.hierarchy.voxels.branch_labels[t]
        grouped_vox_idxs = [
            np.argwhere(voxel_labels == label).flatten()
            for label in np.unique(voxel_labels)
            if label != 0
        ]
        vox_agg = aggregate_stats_for_class(
            self.hierarchy.voxels, t, grouped_vox_idxs, low_memory=self.hierarchy.low_memory
        )
        self.aggregate_voxel_metrics.append(vox_agg)

        if not self.hierarchy.skip_nodes:
            node_labels = self.hierarchy.nodes.branch_label[t]
            grouped_node_idxs = [
                np.argwhere(node_labels == label).flatten()
                for label in np.unique(node_labels)
                if label != 0
            ]
            node_agg = aggregate_stats_for_class(
                self.hierarchy.nodes, t, grouped_node_idxs, low_memory=self.hierarchy.low_memory
            )
            self.aggregate_node_metrics.append(node_agg)

    def _compute_branch_lengths_and_degrees_backend(self, t, xp):
        """
        Compute per-label branch length (centerline length) and neighbor degree per voxel
        using local neighborhood connectivity, with backend xp (np or cp).
        """
        L_cpu = self.hierarchy.im_skel[t]
        L = xp.asarray(L_cpu) if xp is not np else L_cpu
        spacing = self.hierarchy.spacing
        no_z = self.hierarchy.im_info.no_z

        if no_z:
            neighbor_counts = xp.zeros_like(L, dtype=xp.uint8)
            max_label = int(L.max())
            lengths = xp.zeros(max_label + 1, dtype=xp.float32)

            offsets = []
            for dy in (-1, 0, 1):
                for dx in (-1, 0, 1):
                    if dy == 0 and dx == 0:
                        continue
                    if (dy > 0) or (dy == 0 and dx > 0):
                        offsets.append((dy, dx))

            for dy, dx in offsets:
                y0_start = max(0, dy)
                y0_end = L.shape[0] + min(0, dy)
                x0_start = max(0, dx)
                x0_end = L.shape[1] + min(0, dx)

                y1_start = max(0, -dy)
                y1_end = L.shape[0] - max(0, dy)
                x1_start = max(0, -dx)
                x1_end = L.shape[1] - max(0, dx)

                base = L[y0_start:y0_end, x0_start:x0_end]
                neigh = L[y1_start:y1_end, x1_start:x1_end]

                same = (base > 0) & (base == neigh)
                if not xp.any(same):
                    continue

                same_u8 = same.astype(xp.uint8)
                neighbor_counts[y0_start:y0_end, x0_start:x0_end] += same_u8
                neighbor_counts[y1_start:y1_end, x1_start:x1_end] += same_u8

                labels = base[same]
                if labels.size == 0:
                    continue
                dy_phys = dy * spacing[0]
                dx_phys = dx * spacing[1]
                edge_len = xp.sqrt(dy_phys * dy_phys + dx_phys * dx_phys).astype(xp.float32)
                weights = xp.full(labels.size, edge_len, dtype=xp.float32)
                lengths += xp.bincount(labels.ravel(), weights=weights, minlength=max_label + 1)
        else:
            neighbor_counts = xp.zeros_like(L, dtype=xp.uint8)
            max_label = int(L.max())
            lengths = xp.zeros(max_label + 1, dtype=xp.float32)

            offsets = []
            for dz in (-1, 0, 1):
                for dy in (-1, 0, 1):
                    for dx in (-1, 0, 1):
                        if dz == 0 and dy == 0 and dx == 0:
                            continue
                        if (dz > 0) or (dz == 0 and dy > 0) or (dz == 0 and dy == 0 and dx > 0):
                            offsets.append((dz, dy, dx))

            for dz, dy, dx in offsets:
                z0_start = max(0, dz)
                z0_end = L.shape[0] + min(0, dz)
                y0_start = max(0, dy)
                y0_end = L.shape[1] + min(0, dy)
                x0_start = max(0, dx)
                x0_end = L.shape[2] + min(0, dx)

                z1_start = max(0, -dz)
                z1_end = L.shape[0] - max(0, dz)
                y1_start = max(0, -dy)
                y1_end = L.shape[1] - max(0, dy)
                x1_start = max(0, -dx)
                x1_end = L.shape[2] - max(0, dx)

                base = L[z0_start:z0_end, y0_start:y0_end, x0_start:x0_end]
                neigh = L[z1_start:z1_end, y1_start:y1_end, x1_start:x1_end]

                same = (base > 0) & (base == neigh)
                if not xp.any(same):
                    continue

                same_u8 = same.astype(xp.uint8)
                neighbor_counts[z0_start:z0_end, y0_start:y0_end, x0_start:x0_end] += same_u8
                neighbor_counts[z1_start:z1_end, y1_start:y1_end, x1_start:x1_end] += same_u8

                labels = base[same]
                if labels.size == 0:
                    continue
                dz_phys = dz * spacing[0]
                dy_phys = dy * spacing[1]
                dx_phys = dx * spacing[2]
                edge_len = xp.sqrt(
                    dz_phys * dz_phys + dy_phys * dy_phys + dx_phys * dx_phys
                ).astype(xp.float32)
                weights = xp.full(labels.size, edge_len, dtype=xp.float32)
                lengths += xp.bincount(labels.ravel(), weights=weights, minlength=max_label + 1)

        if hasattr(lengths, 'get'):
            lengths_np = lengths.get()
        else:
            lengths_np = np.asarray(lengths)
        if hasattr(neighbor_counts, 'get'):
            neighbor_counts_np = neighbor_counts.get()
        else:
            neighbor_counts_np = np.asarray(neighbor_counts)
        return lengths_np, neighbor_counts_np

    def _compute_branch_lengths_and_degrees(self, t):
        """
        Wrapper that tries GPU backend if enabled, falls back to CPU otherwise.
        """
        if self.hierarchy.use_gpu and _HAS_CUPY:
            try:
                return self._compute_branch_lengths_and_degrees_backend(t, cp)
            except cp.cuda.memory.OutOfMemoryError:
                logger.warning("GPU OOM computing branch lengths; falling back to CPU.")
        return self._compute_branch_lengths_and_degrees_backend(t, np)

    def _get_branch_stats(self, t):
        """
        Compute branch length / thickness / aspect ratio / tortuosity and
        regionprops-based morphology, using an O(N) neighborhood-based algorithm
        rather than an O(N^2) distance matrix.
        """
        branch_idxs_arr = np.array(self.branch_idxs[t])
        L = self.hierarchy.im_skel[t]
        spacing = self.hierarchy.spacing
        no_z = self.hierarchy.im_info.no_z

        # Branch lengths and voxel degrees
        label_lengths, neighbor_counts = self._compute_branch_lengths_and_degrees(t)

        unique_labels = np.unique(L[L > 0])
        if unique_labels.size == 0:
            self.branch_tortuosity.append([])
            self.branch_aspect_ratio.append([])
            self.branch_thickness.append([])
            self.branch_length.append([])
            self.branch_area.append([])
            self.branch_axis_length_maj.append([])
            self.branch_axis_length_min.append([])
            self.branch_extent.append([])
            self.branch_solidity.append([])
            self.reassigned_label.append([])
            self.z.append([])
            self.y.append([])
            self.x.append([])
            return

        # Map neighbor_counts to branch voxel indices
        neighbor_counts_branch = neighbor_counts[tuple(branch_idxs_arr.T)]
        tips = np.where(neighbor_counts_branch == 1)[0]
        lone_tips = np.where(neighbor_counts_branch == 0)[0]

        lone_tip_coords = branch_idxs_arr[lone_tips]
        tip_coords = branch_idxs_arr[tips]

        lone_tip_labels = L[tuple(lone_tip_coords.T)] if len(lone_tip_coords) else np.array([], dtype=int)
        tip_labels = L[tuple(tip_coords.T)] if len(tip_coords) else np.array([], dtype=int)

        # Radii from border
        radii = distance_check(
            self.hierarchy.im_border_mask[t],
            branch_idxs_arr,
            spacing,
        )
        lone_tip_radii = radii[lone_tips] if len(lone_tips) else np.array([], dtype=float)
        tip_radii = radii[tips] if len(tips) else np.array([], dtype=float)

        # Base length per label from adjacency
        base_lengths = np.zeros(len(unique_labels), dtype=np.float32)
        for i, lbl in enumerate(unique_labels):
            if lbl < len(label_lengths):
                base_lengths[i] = label_lengths[int(lbl)]

        # Adjust lengths with tip radii
        for lbl, radius in zip(lone_tip_labels, lone_tip_radii):
            idx = np.where(unique_labels == lbl)[0]
            if idx.size:
                base_lengths[idx[0]] += 2.0 * radius
        for lbl, radius in zip(tip_labels, tip_radii):
            idx = np.where(unique_labels == lbl)[0]
            if idx.size:
                base_lengths[idx[0]] += radius

        # Median thickness per label
        labels_branch_vox = L[tuple(branch_idxs_arr.T)]
        thicknesses = radii * 2.0
        median_thickness = np.zeros(len(unique_labels), dtype=np.float32)
        for i, lbl in enumerate(unique_labels):
            mask = labels_branch_vox == lbl
            if not np.any(mask):
                median_thickness[i] = np.nan
            else:
                median_thickness[i] = np.median(thicknesses[mask])

        # If thickness > length, swap (as in original logic)
        for i in range(len(base_lengths)):
            if not np.isnan(median_thickness[i]) and median_thickness[i] > base_lengths[i]:
                median_thickness[i], base_lengths[i] = base_lengths[i], median_thickness[i]

        aspect_ratios = np.divide(
            base_lengths,
            median_thickness,
            out=np.full_like(base_lengths, np.nan),
            where=median_thickness != 0,
        )

        # Tortuosity: use first two tips per label if available; otherwise 1
        tortuosity = np.ones(len(unique_labels), dtype=np.float32)
        for i, lbl in enumerate(unique_labels):
            mask = tip_labels == lbl
            coords_lbl = tip_coords[mask]
            if coords_lbl.shape[0] >= 2:
                p0, p1 = coords_lbl[0], coords_lbl[1]
                if no_z:
                    dy = (p0[0] - p1[0]) * spacing[0]
                    dx = (p0[1] - p1[1]) * spacing[1]
                    tip_dist = np.sqrt(dy * dy + dx * dx)
                else:
                    dz = (p0[0] - p1[0]) * spacing[0]
                    dy = (p0[1] - p1[1]) * spacing[1]
                    dx = (p0[2] - p1[2]) * spacing[2]
                    tip_dist = np.sqrt(dz * dz + dy * dy + dx * dx)
                if tip_dist > 0:
                    tortuosity[i] = base_lengths[i] / tip_dist
                else:
                    tortuosity[i] = 1.0

        self.branch_tortuosity.append(tortuosity)
        self.branch_aspect_ratio.append(aspect_ratios)
        self.branch_thickness.append(median_thickness)
        self.branch_length.append(base_lengths)

        # regionprops-based morphology and centroids
        regions = regionprops(self.hierarchy.label_branches[t], spacing=self.hierarchy.spacing)
        areas = []
        axis_length_maj = []
        axis_length_min = []
        extent = []
        solidity = []
        reassigned_label = []
        z = []
        y = []
        x = []
        for region in regions:
            reassigned_label_region = np.nan
            if not self.hierarchy.im_info.no_t and self.hierarchy.im_branch_reassigned is not None:
                region_reassigned_labels = self.hierarchy.im_branch_reassigned[t][tuple(region.coords.T)]
                if region_reassigned_labels.size > 0:
                    reassigned_label_region = np.argmax(np.bincount(region_reassigned_labels))
            reassigned_label.append(reassigned_label_region)

            areas.append(region.area)
            try:
                maj_axis = region.major_axis_length
                min_axis = region.minor_axis_length
            except ValueError:
                maj_axis = np.nan
                min_axis = np.nan
            axis_length_maj.append(maj_axis)
            axis_length_min.append(min_axis)
            extent.append(region.extent)
            solidity.append(region.solidity)
            if not self.hierarchy.im_info.no_z:
                z.append(region.centroid[0])
                y.append(region.centroid[1])
                x.append(region.centroid[2])
            else:
                z.append(np.nan)
                y.append(region.centroid[0])
                x.append(region.centroid[1])

        self.branch_area.append(areas)
        self.branch_axis_length_maj.append(axis_length_maj)
        self.branch_axis_length_min.append(axis_length_min)
        self.branch_extent.append(extent)
        self.branch_solidity.append(solidity)
        self.reassigned_label.append(reassigned_label)
        self.z.append(z)
        self.y.append(y)
        self.x.append(x)

    def _run_frame(self, t):
        frame_branch_idxs = np.argwhere(self.hierarchy.im_skel[t] > 0)
        self.branch_idxs.append(frame_branch_idxs)

        frame_skel_branch_labels = self.hierarchy.im_skel[t][tuple(frame_branch_idxs.T)]

        if frame_skel_branch_labels.size == 0:
            # nothing to do
            self.time.append(np.array([], dtype=int))
            self.component_label.append(np.array([], dtype=int))
            self.branch_label.append(np.array([], dtype=int))
            self.image_name.append(np.array([], dtype=object))
            self.aggregate_voxel_metrics.append({})
            if not self.hierarchy.skip_nodes:
                self.aggregate_node_metrics.append({})
            self.branch_length.append([])
            self.branch_thickness.append([])
            self.branch_aspect_ratio.append([])
            self.branch_tortuosity.append([])
            self.branch_area.append([])
            self.branch_axis_length_maj.append([])
            self.branch_axis_length_min.append([])
            self.branch_extent.append([])
            self.branch_solidity.append([])
            self.reassigned_label.append([])
            self.z.append([])
            self.y.append([])
            self.x.append([])
            return

        unique_branch_labels = np.unique(frame_skel_branch_labels)
        unique_branch_labels = unique_branch_labels[unique_branch_labels > 0]
        num_branches = len(unique_branch_labels)

        frame_t = np.ones(num_branches, dtype=int) * t
        self.time.append(frame_t)

        if self.hierarchy.im_info.no_z:
            frame_branch_coords = np.zeros((num_branches, 2), dtype=int)
        else:
            frame_branch_coords = np.zeros((num_branches, 3), dtype=int)

        for idx, lbl in enumerate(unique_branch_labels):
            branch_voxels = frame_branch_idxs[frame_skel_branch_labels == lbl]
            # there should always be at least one voxel for each label here
            if len(branch_voxels) == 0:
                continue
            frame_branch_coords[idx] = branch_voxels[0]

        # Component labels for each branch
        frame_component_label = self.hierarchy.label_components[t][tuple(frame_branch_coords.T)]
        self.component_label.append(frame_component_label)

        # Store the actual branch labels (sorted, unique)
        self.branch_label.append(unique_branch_labels.astype(int))

        im_name = (
            np.ones(num_branches, dtype=object)
            * self.hierarchy.im_info.file_info.filename_no_ext
        )
        self.image_name.append(im_name)

        self._get_aggregate_stats(t)
        self._get_branch_stats(t)

    def run(self):
        for t in range(self.hierarchy.num_t):
            if self.hierarchy.viewer is not None:
                self.hierarchy.viewer.status = (
                    f"Extracting branch features. Frame: {t + 1} of {self.hierarchy.num_t}."
                )
            self._run_frame(t)


class Components:
    """
    Component (organelle)-level features.
    """

    def __init__(self, hierarchy: Hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.component_label = []
        self.aggregate_voxel_metrics = []
        self.aggregate_node_metrics = []
        self.aggregate_branch_metrics = []

        self.z = []
        self.y = []
        self.x = []
        self.organelle_area = []
        self.organelle_axis_length_maj = []
        self.organelle_axis_length_min = []
        self.organelle_extent = []
        self.organelle_solidity = []
        self.reassigned_label = []

        self.image_name = []

        self.stats_to_aggregate = [
            "organelle_area",
            "organelle_axis_length_maj",
            "organelle_axis_length_min",
            "organelle_extent",
            "organelle_solidity",
            "reassigned_label",
        ]

        self.features_to_save = self.stats_to_aggregate + ["x", "y", "z"]

    def _get_aggregate_stats(self, t):
        voxel_labels = self.hierarchy.voxels.component_labels[t]
        grouped_vox_idxs = [
            np.argwhere(voxel_labels == label).flatten()
            for label in np.unique(voxel_labels)
            if label != 0
        ]
        vox_agg = aggregate_stats_for_class(
            self.hierarchy.voxels, t, grouped_vox_idxs, low_memory=self.hierarchy.low_memory
        )
        self.aggregate_voxel_metrics.append(vox_agg)

        if not self.hierarchy.skip_nodes:
            node_labels = self.hierarchy.nodes.component_label[t]
            grouped_node_idxs = [
                np.argwhere(node_labels == label).flatten()
                for label in np.unique(voxel_labels)
                if label != 0
            ]
            node_agg = aggregate_stats_for_class(
                self.hierarchy.nodes, t, grouped_node_idxs, low_memory=self.hierarchy.low_memory
            )
            self.aggregate_node_metrics.append(node_agg)

        branch_labels = self.hierarchy.branches.component_label[t]
        grouped_branch_idxs = [
            np.argwhere(branch_labels == label).flatten()
            for label in np.unique(voxel_labels)
            if label != 0
        ]
        branch_agg = aggregate_stats_for_class(
            self.hierarchy.branches, t, grouped_branch_idxs, low_memory=self.hierarchy.low_memory
        )
        self.aggregate_branch_metrics.append(branch_agg)

    def _get_component_stats(self, t):
        regions = regionprops(self.hierarchy.label_components[t], spacing=self.hierarchy.spacing)
        areas = []
        axis_length_maj = []
        axis_length_min = []
        extent = []
        solidity = []
        reassigned_label = []
        z = []
        y = []
        x = []
        for region in regions:
            reassigned_label_region = np.nan
            if not self.hierarchy.im_info.no_t and self.hierarchy.im_obj_reassigned is not None:
                region_reassigned_labels = self.hierarchy.im_obj_reassigned[t][tuple(region.coords.T)]
                if region_reassigned_labels.size > 0:
                    reassigned_label_region = np.argmax(np.bincount(region_reassigned_labels))
            reassigned_label.append(reassigned_label_region)
            areas.append(region.area)
            try:
                maj_axis = region.major_axis_length
                min_axis = region.minor_axis_length
            except ValueError:
                maj_axis = np.nan
                min_axis = np.nan
            axis_length_maj.append(maj_axis)
            axis_length_min.append(min_axis)
            extent.append(region.extent)
            solidity.append(region.solidity)
            if not self.hierarchy.im_info.no_z:
                z.append(region.centroid[0])
                y.append(region.centroid[1])
                x.append(region.centroid[2])
            else:
                z.append(np.nan)
                y.append(region.centroid[0])
                x.append(region.centroid[1])
        self.organelle_area.append(areas)
        self.organelle_axis_length_maj.append(axis_length_maj)
        self.organelle_axis_length_min.append(axis_length_min)
        self.organelle_extent.append(extent)
        self.organelle_solidity.append(solidity)
        self.reassigned_label.append(reassigned_label)
        self.z.append(z)
        self.y.append(y)
        self.x.append(x)

    def _run_frame(self, t):
        component_labels_t = self.hierarchy.label_components[t]
        mask = component_labels_t > 0
        if not np.any(mask):
            self.component_label.append(np.array([], dtype=int))
            self.time.append(np.array([], dtype=int))
            self.image_name.append(np.array([], dtype=object))
            self.aggregate_voxel_metrics.append({})
            if not self.hierarchy.skip_nodes:
                self.aggregate_node_metrics.append({})
            self.aggregate_branch_metrics.append({})
            self.organelle_area.append([])
            self.organelle_axis_length_maj.append([])
            self.organelle_axis_length_min.append([])
            self.organelle_extent.append([])
            self.organelle_solidity.append([])
            self.reassigned_label.append([])
            self.z.append([])
            self.y.append([])
            self.x.append([])
            return

        frame_component_labels = np.unique(component_labels_t[mask])
        self.component_label.append(frame_component_labels)
        num_components = len(frame_component_labels)

        frame_t = np.ones(num_components, dtype=int) * t
        self.time.append(frame_t)

        im_name = (
            np.ones(num_components, dtype=object)
            * self.hierarchy.im_info.file_info.filename_no_ext
        )
        self.image_name.append(im_name)

        self._get_aggregate_stats(t)
        self._get_component_stats(t)

    def run(self):
        for t in range(self.hierarchy.num_t):
            if self.hierarchy.viewer is not None:
                self.hierarchy.viewer.status = (
                    f"Extracting organelle features. Frame: {t + 1} of {self.hierarchy.num_t}."
                )
            self._run_frame(t)


class Image:
    """
    Image-level aggregated features.
    """

    def __init__(self, hierarchy: Hierarchy):
        self.hierarchy = hierarchy

        self.time = []
        self.image_name = []
        self.aggregate_voxel_metrics = []
        self.aggregate_node_metrics = []
        self.aggregate_branch_metrics = []
        self.aggregate_component_metrics = []
        self.stats_to_aggregate = []
        self.features_to_save = []

    def _get_aggregate_stats(self, t):
        # Voxels: one group containing all voxels in frame t
        voxel_agg = aggregate_stats_for_class(
            self.hierarchy.voxels,
            t,
            [np.arange(len(self.hierarchy.voxels.coords[t]), dtype=int)],
            low_memory=self.hierarchy.low_memory,
        )
        self.aggregate_voxel_metrics.append(voxel_agg)

        # Nodes: one group containing all nodes (if any)
        if not self.hierarchy.skip_nodes:
            node_agg = aggregate_stats_for_class(
                self.hierarchy.nodes,
                t,
                [np.arange(len(self.hierarchy.nodes.nodes[t]), dtype=int)],
                low_memory=self.hierarchy.low_memory,
            )
            self.aggregate_node_metrics.append(node_agg)

        # Branches: one group containing all branches
        n_branches = len(self.hierarchy.branches.branch_length[t])
        branch_idxs = np.arange(n_branches, dtype=int)
        branch_agg = aggregate_stats_for_class(
            self.hierarchy.branches,
            t,
            [branch_idxs],
            low_memory=self.hierarchy.low_memory,
        )
        self.aggregate_branch_metrics.append(branch_agg)

        # Components: one group containing all components
        n_components = len(self.hierarchy.components.organelle_area[t])
        comp_idxs = np.arange(n_components, dtype=int)
        component_agg = aggregate_stats_for_class(
            self.hierarchy.components,
            t,
            [comp_idxs],
            low_memory=self.hierarchy.low_memory,
        )
        self.aggregate_component_metrics.append(component_agg)

    def _run_frame(self, t):
        self.time.append(t)
        self.image_name.append(self.hierarchy.im_info.file_info.filename_no_ext)
        self._get_aggregate_stats(t)

    def run(self):
        for t in range(self.hierarchy.num_t):
            if self.hierarchy.viewer is not None:
                self.hierarchy.viewer.status = (
                    f"Extracting image features. Frame: {t + 1} of {self.hierarchy.num_t}."
                )
            self._run_frame(t)


if __name__ == "__main__":
    im_path = r"F:\60x_568mito_488phal_dapi_siDRP12_w1iSIM-561_s1 - Stage1 _1_-1.tif"
    im_info = ImInfo(im_path)

    hierarchy = Hierarchy(
        im_info,
        skip_nodes=True,
        use_gpu=True,
        low_memory=False,
        enable_motility=True,
        enable_adjacency=True,
    )
    hierarchy.run()



"""
Helpers for adaptive device/memory selection and fallback retries.
"""
from __future__ import annotations

import math
import os


_ESTIMATED_PEAK_MULTIPLIER = 6.0
_MEMORY_HEADROOM = 0.7


def normalize_device(device: str | None) -> str:
    device = (device or "auto").lower()
    if device == "cuda":
        device = "gpu"
    if device not in ("auto", "cpu", "gpu"):
        raise ValueError(f"Unsupported device '{device}'. Use 'auto', 'cpu', or 'gpu'.")
    return device


def gpu_available() -> bool:
    try:
        import cupy
    except ModuleNotFoundError:
        return False
    try:
        return cupy.cuda.runtime.getDeviceCount() > 0
    except Exception:
        return False


def get_gpu_free_bytes() -> int | None:
    try:
        import cupy
    except ModuleNotFoundError:
        return None
    try:
        free_bytes, _ = cupy.cuda.runtime.memGetInfo()
        return int(free_bytes)
    except Exception:
        return None


def _sysconf(name: str) -> int | None:
    try:
        return os.sysconf(name)
    except (AttributeError, ValueError, OSError):
        return None


def get_cpu_available_bytes() -> int | None:
    try:
        import psutil

        return int(psutil.virtual_memory().available)
    except Exception:
        pass

    page_size = _sysconf("SC_PAGE_SIZE") or _sysconf("SC_PAGESIZE")
    if page_size is None:
        return None
    avail_pages = _sysconf("SC_AVPHYS_PAGES")
    if avail_pages is not None:
        return int(avail_pages * page_size)
    total_pages = _sysconf("SC_PHYS_PAGES")
    if total_pages is not None:
        return int(total_pages * page_size)
    return None


def estimate_frame_bytes(im_info) -> int | None:
    if im_info is None or im_info.axes is None or im_info.shape is None:
        return None
    frame_shape = tuple(
        dim for axis, dim in zip(im_info.axes, im_info.shape) if axis != "T"
    )
    if not frame_shape:
        return None
    try:
        itemsize = im_info.im.dtype.itemsize
    except Exception:
        return None
    return int(math.prod(frame_shape) * itemsize)


def should_use_low_memory(im_info, include_gpu: bool) -> bool:
    frame_bytes = estimate_frame_bytes(im_info)
    if frame_bytes is None:
        return False
    peak_bytes = frame_bytes * _ESTIMATED_PEAK_MULTIPLIER
    if include_gpu:
        gpu_free = get_gpu_free_bytes()
        if gpu_free is not None and peak_bytes > gpu_free * _MEMORY_HEADROOM:
            return True
    cpu_free = get_cpu_available_bytes()
    if cpu_free is not None and peak_bytes > cpu_free * _MEMORY_HEADROOM:
        return True
    return False


def mode_candidates(device_order, start_low_memory: bool):
    modes = []
    for dev in device_order:
        modes.append((dev, False))
        modes.append((dev, True))
    if start_low_memory and modes:
        first_device = device_order[0]
        for idx, (dev, low) in enumerate(modes):
            if dev == first_device and low:
                return modes[idx:]
    return modes


def is_oom_error(exc: Exception) -> bool:
    if isinstance(exc, MemoryError):
        return True
    try:
        import cupy

        if isinstance(exc, cupy.cuda.memory.OutOfMemoryError):
            return True
    except Exception:
        pass
    msg = repr(exc).lower()
    return "out of memory" in msg or "outofmemory" in msg


def is_gpu_unavailable_error(exc: Exception) -> bool:
    if isinstance(exc, (ModuleNotFoundError, ImportError)):
        name = getattr(exc, "name", "") or ""
        if "cupy" in name:
            return True
    msg = repr(exc).lower()
    return (
        "gpu backend requested" in msg
        or "cupy is not installed" in msg
        or "cuda is not available" in msg
        or "no cuda devices" in msg
    )



import numpy as np
from types import SimpleNamespace

from nellie.segmentation.mocap_marking import Markers


def _make_im_info(no_z=True):
    if no_z:
        shape = (1, 9, 9)
        axes = "TYX"
        z_res = None
    else:
        shape = (1, 3, 9, 9)
        axes = "TZYX"
        z_res = 0.4
    return SimpleNamespace(
        no_t=True,
        no_z=no_z,
        shape=shape,
        axes=axes,
        dim_res={"X": 0.2, "Y": 0.2, "Z": z_res, "T": 1.0},
    )


def _setup_markers(im_info, intensity, labels, **kwargs):
    markers = Markers(im_info, num_t=1, device="cpu", **kwargs)
    markers.im_memmap = intensity
    markers.label_memmap = labels
    markers.shape = labels.shape
    markers._set_default_sigmas()
    return markers


def test_mocap_marking_low_memory_matches_full_2d():
    im_info = _make_im_info(no_z=True)
    intensity = np.zeros((1, 9, 9), dtype=np.float32)
    intensity[0, 4, 4] = 10.0
    labels = np.zeros((1, 9, 9), dtype=np.uint8)
    labels[0, 2:7, 2:7] = 1

    markers_full = _setup_markers(im_info, intensity, labels, num_sigma=3, low_memory=False)
    full_marker, full_dist, full_border = markers_full._run_frame_impl(0, low_memory=False)

    markers_low = _setup_markers(
        im_info,
        intensity,
        labels,
        num_sigma=3,
        low_memory=True,
        max_chunk_voxels=20,
    )
    low_marker, low_dist, low_border = markers_low._run_frame_impl(
        0, low_memory=True, chunk_voxels=20
    )

    assert np.array_equal(full_marker, low_marker)
    assert np.array_equal(full_dist, low_dist)
    assert np.array_equal(full_border, low_border)


def test_border_is_outside_mask():
    im_info = _make_im_info(no_z=True)
    markers = Markers(im_info, num_t=1, device="cpu")

    mask = np.zeros((7, 7), dtype=bool)
    mask[2:5, 2:5] = True
    _, border = markers._distance_im(mask)

    assert border.shape == mask.shape
    assert not np.any(border & mask)



"""
Example usage and adjacency remapping using voxel matches.

This script avoids allocating large dense voxel-by-voxel matrices and optionally
uses GPU (CuPy) for accumulation when available.
"""
import pickle

import numpy as np

from nellie.im_info.verifier import FileInfo, ImInfo
from nellie.tracking.voxel_reassignment import VoxelReassigner


def accumulate_pair_counts(src_ids, dst_ids, n_src, n_dst, use_gpu=True):
    """
    Accumulate counts into an (n_src, n_dst) matrix given parallel vectors
    src_ids and dst_ids, optionally using cupy if available.
    """
    src_ids = np.asarray(src_ids, dtype=np.int64)
    dst_ids = np.asarray(dst_ids, dtype=np.int64)

    if src_ids.size == 0 or dst_ids.size == 0:
        return np.zeros((n_src, n_dst), dtype=np.uint32)

    if use_gpu:
        try:
            import cupy as cp

            src_gpu = cp.asarray(src_ids)
            dst_gpu = cp.asarray(dst_ids)
            counts_gpu = cp.zeros((n_src, n_dst), dtype=cp.uint32)
            cp.add.at(counts_gpu, (src_gpu, dst_gpu), 1)
            counts = cp.asnumpy(counts_gpu)

            del src_gpu, dst_gpu, counts_gpu
            try:
                cp.get_default_memory_pool().free_all_blocks()
            except Exception:
                pass

            return counts
        except Exception:
            pass

    counts = np.zeros((n_src, n_dst), dtype=np.uint32)
    np.add.at(counts, (src_ids, dst_ids), 1)
    return counts


def main():
    im_path = r"D:\test_files\nelly_smorgasbord\deskewed-iono_pre.ome.tif"
    file_info = FileInfo(im_path)
    file_info.find_metadata()
    file_info.load_metadata()
    im_info = ImInfo(file_info)

    run_obj = VoxelReassigner(im_info, num_t=3)
    run_obj.run()

    edges_loaded = pickle.load(open(im_info.pipeline_paths["adjacency_maps"], "rb"))

    # ------------------------------------------------------------------
    # Example: branch adjacency between t0 and t1 without dense v_t
    # ------------------------------------------------------------------
    mask_01 = run_obj.obj_label_memmap[:2] > 0
    mask_voxels_0 = np.argwhere(mask_01[0])
    mask_voxels_1 = np.argwhere(mask_01[1])

    t0_coords_in_mask_0 = {tuple(coord): idx for idx, coord in enumerate(mask_voxels_0)}
    t1_coords_in_mask_1 = {tuple(coord): idx for idx, coord in enumerate(mask_voxels_1)}

    matches_t0_t1_prev, matches_t0_t1_next = run_obj.running_matches[0]

    idx_matches_0 = []
    idx_matches_1 = []
    for coord_prev, coord_next in zip(matches_t0_t1_prev, matches_t0_t1_next):
        key_prev = tuple(int(c) for c in coord_prev)
        key_next = tuple(int(c) for c in coord_next)
        if key_prev in t0_coords_in_mask_0 and key_next in t1_coords_in_mask_1:
            idx_matches_0.append(t0_coords_in_mask_0[key_prev])
            idx_matches_1.append(t1_coords_in_mask_1[key_next])

    idx_matches_0 = np.asarray(idx_matches_0, dtype=np.int64)
    idx_matches_1 = np.asarray(idx_matches_1, dtype=np.int64)

    b_v0 = edges_loaded["b_v"][0].astype(np.uint8)
    b_v1 = edges_loaded["b_v"][1].astype(np.uint8)

    branch_labels_0 = np.argmax(b_v0, axis=0)
    branch_labels_1 = np.argmax(b_v1, axis=0)

    branch_ids_0 = branch_labels_0[idx_matches_0]
    branch_ids_1 = branch_labels_1[idx_matches_1]

    b0_b1 = accumulate_pair_counts(
        branch_ids_0,
        branch_ids_1,
        n_src=b_v0.shape[0],
        n_dst=b_v1.shape[0],
        use_gpu=True,
    )

    max_idx = np.argmax(b0_b1, axis=0) + 1

    mask_branches = np.zeros(mask_01.shape, dtype=np.uint16)
    mask_branches[0][tuple(mask_voxels_0.T)] = branch_labels_0 + 1

    new_branch_labels_1 = max_idx[branch_labels_1]
    mask_branches[1][tuple(mask_voxels_1.T)] = new_branch_labels_1

    # ------------------------------------------------------------------
    # Example: node adjacency between t0 and t1 (analogous to branches)
    # ------------------------------------------------------------------
    n_v0 = edges_loaded["n_v"][0].astype(np.uint8)
    n_v1 = edges_loaded["n_v"][1].astype(np.uint8)

    node_labels_0 = np.argmax(n_v0, axis=0)
    node_labels_1 = np.argmax(n_v1, axis=0)

    node_ids_0 = node_labels_0[idx_matches_0]
    node_ids_1 = node_labels_1[idx_matches_1]

    n0_n1 = accumulate_pair_counts(
        node_ids_0,
        node_ids_1,
        n_src=n_v0.shape[0],
        n_dst=n_v1.shape[0],
        use_gpu=True,
    )

    max_idx_n = np.argmax(n0_n1, axis=0) + 1

    mask_nodes = np.zeros(mask_01.shape, dtype=np.uint16)
    mask_nodes[0][tuple(mask_voxels_0.T)] = node_labels_0 + 1
    new_node_labels_1 = max_idx_n[node_labels_1]
    mask_nodes[1][tuple(mask_voxels_1.T)] = new_node_labels_1


if __name__ == "__main__":
    main()



from .nellie_processor import NellieProcessor
from .nellie_loader import NellieLoader
from .nellie_analysis import NellieAnalysis
from .nellie_visualizer import NellieVisualizer
from .nellie_settings import Settings
from .nellie_home import Home
from .nellie_fileselect import NellieFileSelect

__all__ = [
    "NellieProcessor",
    "NellieLoader",
    "NellieAnalysis",
    "NellieVisualizer",
    "Settings",
    "Home",
    "NellieFileSelect",
]



"""
Helpers to visualize Hu tracking flow vectors and mocap markers in napari.
"""
from __future__ import annotations

import os
import numpy as np

from nellie.im_info.verifier import ImInfo


def load_flow_vector_array(im_info: ImInfo, path: str | None = None) -> np.ndarray:
    flow_path = path or im_info.pipeline_paths["flow_vector_array"]
    if not os.path.exists(flow_path):
        raise FileNotFoundError(f"Flow vector array not found: {flow_path}")
    return np.load(flow_path)


def flow_vectors_to_tracks(
    flow_vector_array: np.ndarray,
    *,
    no_z: bool,
    cost_threshold: float | None = None,
    stride: int = 1,
    max_vectors: int | None = None,
) -> tuple[np.ndarray, dict]:
    if flow_vector_array.size == 0:
        track_cols = 4 if no_z else 5
        return np.empty((0, track_cols), dtype=np.float32), {"cost": np.array([], dtype=np.float32)}

    flow = flow_vector_array
    if cost_threshold is not None:
        flow = flow[flow[:, -1] <= cost_threshold]
    if stride > 1:
        flow = flow[::stride]
    if max_vectors is not None and flow.shape[0] > max_vectors:
        flow = flow[:max_vectors]

    if flow.size == 0:
        track_cols = 4 if no_z else 5
        return np.empty((0, track_cols), dtype=np.float32), {"cost": np.array([], dtype=np.float32)}

    track_ids = np.arange(flow.shape[0], dtype=np.int64)
    t0 = flow[:, 0].astype(np.int64)
    cost = flow[:, -1].astype(np.float32)

    if no_z:
        y0 = flow[:, 1].astype(np.float32)
        x0 = flow[:, 2].astype(np.float32)
        dy = flow[:, 3].astype(np.float32)
        dx = flow[:, 4].astype(np.float32)
        coords0 = np.column_stack((y0, x0))
        coords1 = np.column_stack((y0 + dy, x0 + dx))
    else:
        z0 = flow[:, 1].astype(np.float32)
        y0 = flow[:, 2].astype(np.float32)
        x0 = flow[:, 3].astype(np.float32)
        dz = flow[:, 4].astype(np.float32)
        dy = flow[:, 5].astype(np.float32)
        dx = flow[:, 6].astype(np.float32)
        coords0 = np.column_stack((z0, y0, x0))
        coords1 = np.column_stack((z0 + dz, y0 + dy, x0 + dx))

    tracks = np.vstack(
        [
            np.column_stack((track_ids, t0, coords0)),
            np.column_stack((track_ids, t0 + 1, coords1)),
        ]
    ).astype(np.float32)
    properties = {"cost": np.repeat(cost, 2)}
    return tracks, properties


def load_flow_vectors_as_tracks(
    im_info: ImInfo,
    *,
    path: str | None = None,
    cost_threshold: float | None = None,
    stride: int = 1,
    max_vectors: int | None = None,
) -> tuple[np.ndarray, dict]:
    flow = load_flow_vector_array(im_info, path=path)
    return flow_vectors_to_tracks(
        flow,
        no_z=im_info.no_z,
        cost_threshold=cost_threshold,
        stride=stride,
        max_vectors=max_vectors,
    )


def load_mocap_markers_as_points(
    im_info: ImInfo,
    *,
    t_range: tuple[int, int] | None = None,
    time_stride: int = 1,
    point_stride: int = 1,
    max_points: int | None = None,
) -> np.ndarray:
    marker_memmap = im_info.get_memmap(im_info.pipeline_paths["im_marker"])
    t_start, t_end = (0, marker_memmap.shape[0]) if t_range is None else t_range

    points = []
    for t in range(t_start, t_end, time_stride):
        coords = np.argwhere(marker_memmap[t] > 0)
        if coords.size == 0:
            continue
        if point_stride > 1:
            coords = coords[::point_stride]
        t_col = np.full((coords.shape[0], 1), t, dtype=np.int64)
        points.append(np.concatenate((t_col, coords.astype(np.int64)), axis=1))

    if points:
        points = np.vstack(points)
    else:
        point_cols = 3 if im_info.no_z else 4
        points = np.empty((0, point_cols), dtype=np.int64)

    if max_points is not None and points.shape[0] > max_points:
        points = points[:max_points]

    return points


if __name__ == "__main__":
    from nellie.im_info.verifier import FileInfo
    
    test_file = "/Users/austin/test_files/nellie_all_tests/yeast_3d_mitochondria.ome_variants/variant_TCZYX_dup2.ome.tif"
    file_info = FileInfo(test_file)
    file_info.find_metadata()
    file_info.load_metadata()

    im_info = ImInfo(file_info)

    tracks, props = load_flow_vectors_as_tracks(im_info, cost_threshold=None, stride=1)
    markers = load_mocap_markers_as_points(im_info, point_stride=1)

    import napari
    viewer = napari.Viewer()
    viewer.add_tracks(tracks, properties=props, name="flow_vectors")
    viewer.add_points(markers, name="mocap_markers", size=2)
    napari.run()



"""
Label tracking across timepoints using flow interpolation.

This module provides the LabelTracks class for tracking labeled objects over time
using forward and backward flow interpolation.
"""
import numpy as np

from nellie.im_info.verifier import ImInfo
from nellie.tracking.flow_interpolation import interpolate_all_forward, interpolate_all_backward


class LabelTracks:
    """
    A class to track labeled objects over multiple timepoints in a microscopy image using flow interpolation.

    Attributes
    ----------
    im_info : ImInfo
        An object containing image metadata and memory-mapped image data.
    num_t : int
        Number of timepoints in the image.
    label_im_path : str
        Path to the labeled instance image.
    im_memmap : np.ndarray or None
        Memory-mapped original image data.
    label_memmap : np.ndarray or None
        Memory-mapped labeled instance image data.

    Methods
    -------
    initialize()
        Initializes memory-mapped data for both the raw image and the labeled instance image.
    run(label_num=None, start_frame=0, end_frame=None, min_track_num=0, skip_coords=1, max_distance_um=0.5)
        Runs the tracking process for labeled objects across timepoints, both forward and backward.
    """
    def __init__(self, im_info: ImInfo, num_t: int = None, label_im_path: str = None):
        """
        Initializes the LabelTracks class with image metadata, timepoints, and label image path.

        Parameters
        ----------
        im_info : ImInfo
            An instance of the ImInfo class containing image metadata and paths.
        num_t : int, optional
            Number of timepoints in the image (default is None, in which case it is inferred from the image metadata).
        label_im_path : str, optional
            Path to the labeled instance image. If not provided, defaults to the 'im_instance_label' path in `im_info`.
        """
        self.im_info = im_info
        self.num_t = num_t
        if label_im_path is None:
            label_im_path = self.im_info.pipeline_paths['im_instance_label']
        self.label_im_path = label_im_path

        if num_t is None:
            self.num_t = im_info.shape[im_info.axes.index('T')]

        self.im_memmap = None
        self.label_memmap = None

    def initialize(self):
        """
        Initializes memory-mapped data for both the raw image and the labeled instance image.

        This method prepares the image data and the labeled data for processing, mapping them into memory.
        """
        self.label_memmap = self.im_info.get_memmap(self.label_im_path)
        self.im_memmap = self.im_info.get_memmap(self.im_info.im_path)

    def run(self, label_num=None, start_frame=0, end_frame=None, min_track_num=0, skip_coords=1, max_distance_um=0.5):
        """
        Runs the tracking process for labeled objects across timepoints, using flow interpolation.

        This method uses forward and backward interpolation to track objects across multiple frames, starting from a given
        frame. It can also track specific labels or all labels in the image.

        Parameters
        ----------
        label_num : int, optional
            Label number to track. If None, all labels are tracked (default is None).
        start_frame : int, optional
            The starting frame from which to begin tracking (default is 0).
        end_frame : int, optional
            The ending frame for the tracking. If None, tracks until the last frame (default is None).
        min_track_num : int, optional
            Minimum track number to assign to the coordinates (default is 0).
        skip_coords : int, optional
            The interval at which coordinates are sampled (default is 1).
        max_distance_um : float, optional
            Maximum distance allowed for interpolation (in micrometers, default is 0.5).

        Returns
        -------
        tuple
            A list of tracks and a dictionary of track properties.
        """
        if end_frame is None:
            end_frame = self.num_t
        num_frames = self.label_memmap.shape[0] - 1
        if start_frame > num_frames:
            return [], {}
        if label_num is None:
            coords = np.argwhere(self.label_memmap[start_frame] > 0).astype(float)
        else:
            coords = np.argwhere(self.label_memmap[start_frame] == label_num).astype(float)
        if coords.shape[0] == 0:
            return [], {}
        coords = np.array(coords[::skip_coords])
        coords_copy = coords.copy()
        tracks = []
        track_properties = {}
        if start_frame < end_frame:
            tracks, track_properties = interpolate_all_forward(coords, start_frame, end_frame, self.im_info,
                                                               min_track_num, max_distance_um=max_distance_um)

        new_end_frame = 0  # max(0, end_frame - start_frame)
        if start_frame > 0:
            tracks_bw, track_properties_bw = interpolate_all_backward(coords_copy, start_frame, new_end_frame,
                                                                      self.im_info, min_track_num,
                                                                      max_distance_um=max_distance_um)
            tracks_bw = tracks_bw[::-1]
            for property in track_properties_bw.keys():
                track_properties_bw[property] = track_properties_bw[property][::-1]
            sort_idx = np.argsort([track[0] for track in tracks_bw])
            tracks_bw = [tracks_bw[i] for i in sort_idx]
            tracks = tracks_bw + tracks
            for property in track_properties_bw.keys():
                track_properties_bw[property] = [track_properties_bw[property][i] for i in sort_idx]
            if not track_properties:
                track_properties = track_properties_bw
            else:
                for property in track_properties_bw.keys():
                    track_properties[property] = track_properties_bw[property] + track_properties[property]

        # Let's get rid of any tracks that are > 1 pixel away from the nearest mask pixel
        filtered_tracks = []
        filtered_track_properties = {key: [] for key in track_properties} if track_properties else {}

        for track_num, track in enumerate(tracks):
            dims = tuple(int(np.round(d)) for d in track[1:])
            is_in_range = True
            for i, dim_val in enumerate(dims):
                if not (0 <= dim_val < self.label_memmap.shape[i]):
                    is_in_range = False
                    break
            # Check if the track is valid (not too far from a mask pixel and within bounds)
            if is_in_range and np.min(self.label_memmap[dims]) > 0:
                filtered_tracks.append(track)
                for prop_key, prop_list in track_properties.items():
                    filtered_track_properties[prop_key].append(prop_list[track_num])

        tracks = filtered_tracks
        track_properties = filtered_track_properties
                
        return tracks, track_properties


if __name__ == "__main__":
    im_path = r"D:\test_files\nellie_longer_smorgasbord\deskewed-peroxisome.ome.tif"
    im_info = ImInfo(im_path, ch=0)
    num_t = 20
    label_tracks = LabelTracks(im_info, num_t=num_t)
    label_tracks.initialize()
    # tracks, track_properties = label_tracks.run(label_num=None, skip_coords=1)

    all_tracks = []
    all_props = {}
    max_track_num = 0
    for frame in range(num_t):
        tracks, track_properties = label_tracks.run(label_num=None, start_frame=frame, end_frame=None,
                                                    min_track_num=max_track_num,
                                                    skip_coords=100)
        all_tracks += tracks
        for property in track_properties.keys():
            if property not in all_props.keys():
                all_props[property] = []
            all_props[property] += track_properties[property]
        if len(tracks) == 0:
            break
        max_track_num = max([track[0] for track in tracks]) + 1

    # pickle tracks and track properties
    import pickle
    import datetime
    dt = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    with open(f'{dt}-mt_tracks.pkl', 'wb') as f:
        pickle.dump(all_tracks, f)
    with open(f'{dt}-mt_props.pkl', 'wb') as f:
        pickle.dump(all_props, f)

    # import napari
    # viewer = napari.Viewer()
    #
    # raw_im = im_info.get_im_memmap(im_info.im_path)[:num_t]
    # viewer.add_image(raw_im, name='raw_im')
    # viewer.add_tracks(all_tracks, properties=all_props, name='tracks')
    # napari.run()
